@@ -162,10 +162,15 @@ struct acmp_send_queue {
 struct acmp_addr {
 	uint16_t              type;
 	union acm_ep_info     info;
-	struct acm_address    *addr;
+	struct acm_address    addr;
 	struct acmp_ep        *ep;
 };
 
+struct acmp_addr_ctx {
+	struct acmp_ep	     *ep;
+	int		     addr_inx;
+};
+
 struct acmp_ep {
 	struct acmp_port      *port;
 	struct ibv_cq         *cq;
@@ -186,7 +191,10 @@ struct acmp_ep {
 	struct list_head      active_queue;
 	struct list_head      wait_queue;
 	enum acmp_state       state;
-	struct acmp_addr      addr_info[MAX_EP_ADDR];
+	/* This lock protects nmbr_ep_addrs and addr_info */
+	pthread_rwlock_t      rwlock;
+	int		      nmbr_ep_addrs;
+	struct acmp_addr      *addr_info;
 	atomic_t              counters[ACM_MAX_COUNTER];
 };
 
@@ -1040,9 +1048,11 @@ acmp_resolve_sa_resp(struct acm_sa_mad *mad)
 static struct acmp_addr *
 acmp_addr_lookup(struct acmp_ep *ep, uint8_t *addr, uint16_t type)
 {
+	struct acmp_addr *ret = NULL;
 	int i;
 
-	for (i = 0; i < MAX_EP_ADDR; i++) {
+	pthread_rwlock_rdlock(&ep->rwlock);
+	for (i = 0; i < ep->nmbr_ep_addrs; i++) {
 		if (ep->addr_info[i].type != type)
 			continue;
 
@@ -1051,11 +1061,13 @@ acmp_addr_lookup(struct acmp_ep *ep, uint8_t *addr, uint16_t type)
 			      (char *) addr, ACM_MAX_ADDRESS)) ||
 		    !memcmp(ep->addr_info[i].info.addr, addr,
 			    ACM_MAX_ADDRESS)) {
-			return &ep->addr_info[i];
+			ret = ep->addr_info + i;
+			break;
 		}
 	}
+	pthread_rwlock_unlock(&ep->rwlock);
 
-	return NULL;
+	return ret;
 }
 
 static void
@@ -1593,13 +1605,12 @@ static void *acmp_retry_handler(void *context)
 	return NULL;
 }
 
+/* rwlock must be held read-locked */
 static int
-acmp_query(void *addr_context, struct acm_msg *msg, uint64_t id)
+__acmp_query(struct acmp_ep *ep, struct acm_msg *msg, uint64_t id)
 {
 	struct acmp_request *req;
 	struct ib_sa_mad *mad;
-	struct acmp_addr *address = addr_context;
-	struct acmp_ep *ep = address->ep;
 	uint8_t status;
 	struct acm_sa_mad *sa_mad;
 
@@ -1652,6 +1663,21 @@ resp:
 	return acm_query_response(id, msg);
 }
 
+static int
+acmp_query(void *addr_context, struct acm_msg *msg, uint64_t id)
+{
+	struct acmp_addr_ctx *addr_ctx = addr_context;
+	struct acmp_addr *address;
+	int ret;
+
+	pthread_rwlock_rdlock(&addr_ctx->ep->rwlock);
+	address = addr_ctx->ep->addr_info + addr_ctx->addr_inx;
+	ret = __acmp_query(address->ep, msg, id);
+	pthread_rwlock_unlock(&addr_ctx->ep->rwlock);
+
+	return ret;
+}
+
 static uint8_t
 acmp_send_resolve(struct acmp_ep *ep, struct acmp_dest *dest,
 	struct acm_ep_addr_data *saddr)
@@ -1940,7 +1966,8 @@ put:
 static int
 acmp_resolve(void *addr_context, struct acm_msg *msg, uint64_t id)
 {
-	struct acmp_addr *address = addr_context;
+	struct acmp_addr_ctx *addr_ctx = addr_context;
+	struct acmp_addr *address = addr_ctx->ep->addr_info + addr_ctx->addr_inx;
 	struct acmp_ep *ep = address->ep;
 
 	if (ep->state != ACMP_READY) {
@@ -2356,34 +2383,50 @@ static void acmp_ep_preload(struct acmp_ep *ep)
 	}
 }
 
-static int acmp_add_addr(const struct acm_address *addr, void *ep_context,
-			 void **addr_context)
+/* rwlock must be held write-locked */
+static int __acmp_add_addr(const struct acm_address *addr, struct acmp_ep *ep,
+			   void **addr_context)
 {
-	struct acmp_ep *ep = ep_context;
 	struct acmp_dest *dest;
+	struct acmp_addr_ctx *addr_ctx;
 	int i;
 
-	acm_log(2, "\n");
-
-	for (i = 0; (i < MAX_EP_ADDR) &&
+	for (i = 0; (i < ep->nmbr_ep_addrs) &&
 	     (ep->addr_info[i].type != ACM_ADDRESS_INVALID); i++)
 		;
 
-	if (i == MAX_EP_ADDR) {
-		acm_log(0, "ERROR - no more space for local address\n");
-		return -1;
+	if (i == ep->nmbr_ep_addrs) {
+		struct acmp_addr *new_info;
+
+		new_info = realloc(ep->addr_info, (i + 1) * sizeof(*ep->addr_info));
+		if (!new_info) {
+			acm_log(0, "ERROR - no more space for local address\n");
+			return -1;
+		}
+		ep->addr_info = new_info;
+		/* Added memory is not initialized */
+		memset(ep->addr_info + i, 0, sizeof(*ep->addr_info));
+		++ep->nmbr_ep_addrs;
 	}
 	ep->addr_info[i].type = addr->type;
 	memcpy(&ep->addr_info[i].info, &addr->info, sizeof(addr->info));
-	ep->addr_info[i].addr = (struct acm_address *) addr;
+	memcpy(&ep->addr_info[i].addr, addr, sizeof(*addr));
 	ep->addr_info[i].ep = ep;
 
+	addr_ctx = malloc(sizeof(*addr_ctx));
+	if (!addr_ctx) {
+		acm_log(0, "ERROR - unable to alloc address context struct\n");
+		return -1;
+	}
+	addr_ctx->ep = ep;
+	addr_ctx->addr_inx = i;
+
 	if (loopback_prot != ACMP_LOOPBACK_PROT_LOCAL) {
-		*addr_context = &ep->addr_info[i];
+		*addr_context = addr_ctx;
 		return 0;
 	}
 
-	dest = acmp_acquire_dest(ep, addr->type, (uint8_t *) addr->info.addr);
+	dest = acmp_acquire_dest(ep, addr->type, (uint8_t *)addr->info.addr);
 	if (!dest) {
 		acm_log(0, "ERROR - unable to create loopback dest %s\n",
 			addr->id_string);
@@ -2404,15 +2447,31 @@ static int acmp_add_addr(const struct acm_address *addr, void *ep_context,
 	dest->route_timeout = (uint64_t) ~0ULL;
 	dest->state = ACMP_READY;
 	acmp_put_dest(dest);
-	*addr_context = &ep->addr_info[i];
+	*addr_context = addr_ctx;
 	acm_log(1, "added loopback dest %s\n", dest->name);
 
 	return 0;
 }
 
+static int acmp_add_addr(const struct acm_address *addr, void *ep_context,
+			 void **addr_context)
+{
+	struct acmp_ep *ep = ep_context;
+	int ret;
+
+	acm_log(2, "\n");
+
+	pthread_rwlock_wrlock(&ep->rwlock);
+	ret = __acmp_add_addr(addr, ep, addr_context);
+	pthread_rwlock_unlock(&ep->rwlock);
+
+	return ret;
+}
+
 static void acmp_remove_addr(void *addr_context)
 {
-	struct acmp_addr *address = addr_context;
+	struct acmp_addr_ctx *addr_ctx = addr_context;
+	struct acmp_addr *address = addr_ctx->ep->addr_info + addr_ctx->addr_inx;
 	struct acmp_device *dev;
 	struct acmp_dest *dest;
 	struct acmp_ep *ep;
@@ -2435,7 +2494,7 @@ static void acmp_remove_addr(void *addr_context)
 			pthread_mutex_lock(&port->lock);
 			list_for_each(&port->ep_list, ep, entry) {
 				pthread_mutex_unlock(&port->lock);
-				dest = acmp_get_dest(ep, address->type, address->addr->info.addr);
+				dest = acmp_get_dest(ep, address->type, address->addr.info.addr);
 				if (dest) {
 					acm_log(2, "Found a dest addr, deleting it\n");
 					pthread_mutex_lock(&ep->lock);
@@ -2451,6 +2510,7 @@ static void acmp_remove_addr(void *addr_context)
 	pthread_mutex_unlock(&acmp_dev_lock);
 
 	memset(address, 0, sizeof(*address));
+	free(addr_ctx);
 }
 
 static struct acmp_port *acmp_get_port(struct acm_endpoint *endpoint)
@@ -2536,6 +2596,14 @@ acmp_alloc_ep(struct acmp_port *port, struct acm_endpoint *endpoint)
 	pthread_mutex_init(&ep->lock, NULL);
 	sprintf(ep->id_string, "%s-%d-0x%x", port->dev->verbs->device->name,
 		port->port_num, endpoint->pkey);
+
+	if (pthread_rwlock_init(&ep->rwlock, NULL)) {
+		free(ep);
+		return NULL;
+	}
+	ep->addr_info = NULL;
+	ep->nmbr_ep_addrs = 0;
+
 	for (i = 0; i < ACM_MAX_COUNTER; i++)
 		atomic_init(&ep->counters[i]);
 
