@@ -5,10 +5,13 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <stdbool.h>
 #include <errno.h>
 #include <unistd.h>
+#include <getopt.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <syslog.h>
 #include <rdma/rdma_netlink.h>
 #include <netlink/netlink.h>
 #include <netlink/msg.h>
@@ -22,7 +25,7 @@
  *                 by->onboard -> by-pci -> by-guid -> kernel
  * NAME_KERNEL - leave name as kernel provided
  * NAME_PCI - based on PCI/slot/function location
- * NAME_GUID - based on system image GUID
+ * NAME_GUID - based on node GUID
  * NAME_ONBOARD - based on-board device index
  *
  * The stable names are combination of device type technology and rename mode.
@@ -30,6 +33,7 @@
  * RoCE - roce*
  * iWARP - iw*
  * OPA - opa*
+ * Default (unknown protocol) - rdma*
  *
  * Example:
  * NAME_PCI
@@ -49,11 +53,19 @@
 struct data {
 	const char *curr;
 	char *prefix;
-	uint64_t sys_image_guid;
+	uint64_t node_guid;
 	char *name;
 	int idx;
 };
 
+static bool debug_mode;
+#define pr_err(args...) syslog(LOG_ERR, ##args)
+#define pr_dbg(args...)                                                        \
+	do {                                                                   \
+		if (debug_mode)                                                \
+			syslog(LOG_ERR, ##args);                               \
+	} while (0)
+
 #define ONBOARD_INDEX_MAX (16*1024-1)
 static int by_onboard(struct data *d)
 {
@@ -84,6 +96,7 @@ static int by_onboard(struct data *d)
 	if (!fp)
 		fp = fopen(index, "r");
 	if (!fp) {
+		pr_dbg("%s: Device is not embedded onboard\n", d->curr);
 		ret = -ENOENT;
 		goto out;
 	}
@@ -92,12 +105,15 @@ static int by_onboard(struct data *d)
 	fclose(fp);
 	/* https://github.com/systemd/systemd/blob/master/src/udev/udev-builtin-net_id.c#L263 */
 	if (!ret || o > ONBOARD_INDEX_MAX) {
+		pr_err("%s: Onboard index %d and ret %d\n", d->curr, o, ret);
 		ret = -ENOENT;
 		goto out;
 	}
 
 	ret = asprintf(&d->name, "%so%u", d->prefix, o);
 	if (ret < 0) {
+		pr_err("%s: Failed to allocate name with prefix %s and onboard index %d\n",
+		       d->curr, d->prefix, o);
 		ret = -ENOENT;
 		d->name = NULL;
 	}
@@ -152,7 +168,7 @@ static int find_sun(char *devname, char *pci)
 			goto out;
 		}
 
-		ret = fscanf(fp, "%s", address);
+		ret = fscanf(fp, "%255s", address);
 		fclose(fp);
 
 		if (ret != 1) {
@@ -217,13 +233,133 @@ static int is_pci_ari_enabled(char *devname)
 	return (ret) ? a == 1 : 0;
 }
 
+struct pci_info {
+	char *pcidev;
+
+	unsigned int domain;
+	unsigned int bus;
+	unsigned int slot;
+	unsigned int func;
+	unsigned int sun;
+	unsigned int vf;
+	bool valid_vf;
+};
+
+static int fill_pci_info(struct data *d, struct pci_info *p)
+{
+	char buf[256] = {};
+	char *pci;
+	int ret;
+
+	ret = readlink(p->pcidev, buf, sizeof(buf)-1);
+	if (ret == -1 || ret == sizeof(buf))
+		return -EINVAL;
+
+	buf[ret] = 0;
+
+	pci = basename(buf);
+	/*
+	 * pci = 0000:00:0c.0
+	 */
+	ret = sscanf(pci, "%x:%x:%x.%u", &p->domain, &p->bus, &p->slot,
+		     &p->func);
+	if (ret != 4) {
+		pr_err("%s: Failed to read PCI BOF\n", d->curr);
+		return -ENOENT;
+	}
+
+	if (is_pci_ari_enabled(p->pcidev)) {
+		/*
+		 * ARI devices support up to 256 functions on a single device
+		 * ("slot"), and interpret the traditional 5-bit slot and 3-bit
+		 * function number as a single 8-bit function number, where the
+		 * slot makes up the upper 5 bits.
+		 *
+		 * https://github.com/systemd/systemd/blob/master/src/udev/udev-builtin-net_id.c#L344
+		 */
+		p->func += p->slot * 8;
+		pr_dbg("%s: This is ARI device, new PCI BOF is %04x:%02x:%02x.%u\n",
+		       d->curr, p->domain, p->bus, p->slot, p->func);
+	}
+
+	p->sun = find_sun(p->pcidev, pci);
+
+	return 0;
+}
+
+static int get_virtfn_info(struct data *d, struct pci_info *p)
+{
+	struct pci_info vf = {};
+	char *physfn_pcidev;
+	struct dirent *dent;
+	DIR *dir;
+	int ret;
+
+	/* Check if this is a virtual function. */
+	ret = asprintf(&physfn_pcidev, "%s/physfn", p->pcidev);
+	if (ret < 0)
+		return -ENOMEM;
+
+	/* We are VF, get VF number and replace pcidev to point to PF */
+	dir = opendir(physfn_pcidev);
+	if (!dir) {
+		/*
+		 * -ENOENT means that we are already in PF
+		 *  and pcidev points to right PCI.
+		 */
+		ret = (errno == ENOENT) ? 0 : -ENOMEM;
+		goto err_free;
+	}
+
+	p->valid_vf = true;
+	vf.pcidev = p->pcidev;
+	ret = fill_pci_info(d, &vf);
+	if (ret)
+		goto err_dir;
+
+	while ((dent = readdir(dir))) {
+		const char *s = "virtfn";
+		struct pci_info v = {};
+
+		if (strncmp(dent->d_name, s, strlen(s)) ||
+		    strlen(dent->d_name) == strlen(s))
+			continue;
+
+		ret = asprintf(&v.pcidev, "%s/%s", physfn_pcidev, dent->d_name);
+		if (ret < 0) {
+			ret = -ENOMEM;
+			goto err_dir;
+		}
+		ret = fill_pci_info(d, &v);
+		free(v.pcidev);
+		if (ret) {
+			ret = -ENOMEM;
+			goto err_dir;
+		}
+		if (vf.func == v.func) {
+			p->vf = atoi(&dent->d_name[6]);
+			break;
+		}
+	}
+
+	p->pcidev = physfn_pcidev;
+	closedir(dir);
+
+	return 0;
+
+err_dir:
+	closedir(dir);
+err_free:
+	free(physfn_pcidev);
+	return ret;
+}
+
 static int by_pci(struct data *d)
 {
-	long domain, bus, slot, func, sun;
-	char *devpath = NULL;
+	struct pci_info p = {};
 	char *subsystem;
 	char buf[256] = {};
-	char *pci, *subs;
+	char *subs;
 	int ret;
 
 	ret = asprintf(&subsystem, "/sys/class/infiniband/%s/device/subsystem",
@@ -241,43 +377,28 @@ static int by_pci(struct data *d)
 	subs = basename(buf);
 	if (strcmp(subs, "pci")) {
 		/* Ball out virtual devices */
+		pr_dbg("%s: Non-PCI device (%s) was detected\n", d->curr, subs);
 		ret = -EINVAL;
 		goto out;
 	}
+
 	/* Real devices */
-	ret = asprintf(&devpath, "/sys/class/infiniband/%s/device", d->curr);
+	ret = asprintf(&p.pcidev, "/sys/class/infiniband/%s/device", d->curr);
 	if (ret < 0) {
 		ret = -ENOMEM;
-		devpath = NULL;
+		p.pcidev = NULL;
 		goto out;
 	}
 
-	ret = readlink(devpath, buf, sizeof(buf)-1);
-	if (ret == -1 || ret == sizeof(buf)) {
-		ret = -EINVAL;
+	ret = get_virtfn_info(d, &p);
+	if (ret)
+		goto out;
+
+	ret = fill_pci_info(d, &p);
+	if (ret) {
+		pr_err("%s: Failed to fill PCI device information\n", d->curr);
 		goto out;
 	}
-	buf[ret] = 0;
-
-	pci = basename(buf);
-	/*
-	 * pci = 0000:00:0c.0
-	 */
-	if (sscanf(pci, "%lx:%lx:%lx.%lu", &domain, &bus, &slot, &func) != 4) {
-		ret = -ENOENT;
-		goto out;
-	}
-
-	if (is_pci_ari_enabled(devpath))
-		/*
-		 * ARI devices support up to 256 functions on a single device
-		 * ("slot"), and interpret the traditional 5-bit slot and 3-bit
-		 * function number as a single 8-bit function number, where the
-		 * slot makes up the upper 5 bits.
-		 *
-		 * https://github.com/systemd/systemd/blob/master/src/udev/udev-builtin-net_id.c#L344
-		 */
-		func += slot * 8;
 
 	d->name = calloc(256, sizeof(char));
 	if (!d->name) {
@@ -291,8 +412,8 @@ static int by_pci(struct data *d)
 		goto out;
 	}
 
-	if (domain > 0) {
-		ret = sprintf(buf, "P%ld", domain);
+	if (p.domain > 0) {
+		ret = sprintf(buf, "P%u", p.domain);
 		if (ret == -1) {
 			ret = -ENOMEM;
 			goto out;
@@ -300,11 +421,10 @@ static int by_pci(struct data *d)
 		strcat(d->name, buf);
 	}
 
-	sun = find_sun(devpath, pci);
-	if (sun > 0)
-		ret = sprintf(buf, "s%ld", sun);
+	if (p.sun > 0)
+		ret = sprintf(buf, "s%u", p.sun);
 	else
-		ret = sprintf(buf, "p%lds%ld", bus, slot);
+		ret = sprintf(buf, "p%us%u", p.bus, p.slot);
 	if (ret == -1) {
 		ret = -ENOMEM;
 		goto out;
@@ -312,17 +432,26 @@ static int by_pci(struct data *d)
 
 	strcat(d->name, buf);
 
-	if (func > 0 || is_pci_multifunction(devpath)) {
-		ret = sprintf(buf, "f%ld", func);
+	if (p.func > 0 || is_pci_multifunction(p.pcidev)) {
+		ret = sprintf(buf, "f%u", p.func);
 		if (ret == -1) {
 			ret = -ENOMEM;
 			goto out;
 		}
 		strcat(d->name, buf);
+
+		if (p.valid_vf) {
+			ret = sprintf(buf, "v%u", p.vf);
+			if (ret == -1) {
+				ret = -ENOMEM;
+				goto out;
+			}
+			strcat(d->name, buf);
+		}
 	}
 	ret = 0;
 out:
-	free(devpath);
+	free(p.pcidev);
 	free(subsystem);
 	if (ret) {
 		free(d->name);
@@ -337,11 +466,11 @@ static int by_guid(struct data *d)
 	uint16_t vp[4];
 	int ret = -1;
 
-	if (!d->sys_image_guid)
+	if (!d->node_guid)
 		/* virtual devices start without GUID */
 		goto out;
 
-	memcpy(vp, &d->sys_image_guid, sizeof(uint64_t));
+	memcpy(vp, &d->node_guid, sizeof(uint64_t));
 	ret = asprintf(&d->name, "%sx%04x%04x%04x%04x", d->prefix, vp[3], vp[2],
 		       vp[1], vp[0]);
 out:
@@ -393,17 +522,23 @@ static int get_nldata_cb(struct nl_msg *msg, void *data)
 		return NL_STOP;
 
 	if (!tb[RDMA_NLDEV_ATTR_DEV_NAME] || !tb[RDMA_NLDEV_ATTR_DEV_INDEX] ||
-	    !tb[RDMA_NLDEV_ATTR_SYS_IMAGE_GUID] ||
-	    !tb[RDMA_NLDEV_ATTR_DEV_PROTOCOL])
+	    !tb[RDMA_NLDEV_ATTR_NODE_GUID])
 		return NL_STOP;
 
 	ret = strcmp(d->curr, nla_get_string(tb[RDMA_NLDEV_ATTR_DEV_NAME]));
 	if (ret)
 		return NL_OK;
 
-	d->prefix = strdup(nla_get_string(tb[RDMA_NLDEV_ATTR_DEV_PROTOCOL]));
+	if (tb[RDMA_NLDEV_ATTR_DEV_PROTOCOL])
+		d->prefix = strdup(
+			nla_get_string(tb[RDMA_NLDEV_ATTR_DEV_PROTOCOL]));
+	if (!d->prefix)
+		ret = asprintf(&d->prefix, "rdma");
+	if (ret < 0)
+		return NL_STOP;
+
 	d->idx = nla_get_u32(tb[RDMA_NLDEV_ATTR_DEV_INDEX]);
-	d->sys_image_guid = nla_get_u64(tb[RDMA_NLDEV_ATTR_SYS_IMAGE_GUID]);
+	d->node_guid = nla_get_u64(tb[RDMA_NLDEV_ATTR_NODE_GUID]);
 	return NL_STOP;
 }
 
@@ -430,32 +565,61 @@ static int str2policy(const char *np)
 	return NAME_ERROR;
 };
 
-int main(int argc, const char *argv[])
+int main(int argc, char **argv)
 {
 	struct data d = { .idx = -1 };
 	struct nl_sock *nl;
 	int ret = -1;
-	int np;
+	int np, opt;
 
-	if (argc != 3)
+	if (argc < 3)
 		goto err;
 
-	np = str2policy(argv[2]);
-	if (np & NAME_ERROR)
-		goto err;
+	while ((opt = getopt(argc, argv, "v")) >= 0) {
+		switch (opt) {
+		case 'v':
+			debug_mode = true;
+			break;
+		default:
+			goto err;
+		}
+	}
 
-	if (np & NAME_KERNEL)
+	argc -= optind;
+	argv += optind;
+
+	d.curr = argv[0];
+
+	np = str2policy(argv[1]);
+	if (np & NAME_ERROR) {
+		pr_err("%s: Unknown policy %s\n", d.curr, argv[1]);
+		goto err;
+	}
+
+	pr_dbg("%s: Requested policy is %s\n", d.curr, argv[1]);
+
+	if (np & NAME_KERNEL) {
+		pr_dbg("%s: Leave kernel names, do nothing\n", d.curr);
 		/* Do nothing */
 		exit(0);
+	}
 
 	nl = rdmanl_socket_alloc();
-	if (!nl)
+	if (!nl) {
+		pr_err("%s: Failed to allocate netlink socket\n", d.curr);
 		goto err;
+	}
 
-	d.curr = argv[1];
-	if (rdmanl_get_devices(nl, get_nldata_cb, &d) || d.idx == -1 ||
-	    !d.prefix)
+	if (rdmanl_get_devices(nl, get_nldata_cb, &d)) {
+		pr_err("%s: Failed to connect to NETLINK_RDMA\n", d.curr);
 		goto out;
+	}
+
+	if (d.idx == -1 || !d.prefix) {
+		pr_err("%s: Failed to get current device name and index\n",
+		       d.curr);
+		goto out;
+	}
 
 	ret = -1;
 	if (np & NAME_ONBOARD)
@@ -468,8 +632,12 @@ int main(int argc, const char *argv[])
 		goto out;
 
 	ret = device_rename(nl, &d);
-	if (ret)
+	if (ret) {
+		pr_err("%s: Device rename to %s failed with error %d\n", d.curr,
+		       d.name, ret);
 		goto out;
+	}
+	pr_dbg("%s: Successfully renamed device to be %s\n", d.curr, d.name);
 
 	printf("%s\n", d.name);
 	free(d.name);
