@@ -39,25 +39,25 @@ static int advise_mr(struct ibv_pd *pd,
 		     struct ibv_sge *sg_list,
 		     uint32_t num_sges)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static struct ibv_dm *alloc_dm(struct ibv_context *context,
 			       struct ibv_alloc_dm_attr *attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_mw *alloc_mw(struct ibv_pd *pd, enum ibv_mw_type type)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_mr *alloc_null_mr(struct ibv_pd *pd)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
@@ -65,20 +65,20 @@ static struct ibv_pd *
 alloc_parent_domain(struct ibv_context *context,
 		    struct ibv_parent_domain_init_attr *attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_pd *alloc_pd(struct ibv_context *context)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_td *alloc_td(struct ibv_context *context,
 			       struct ibv_td_init_attr *init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
@@ -91,24 +91,24 @@ static int attach_counters_point_flow(struct ibv_counters *counters,
 				      struct ibv_counter_attach_attr *attr,
 				      struct ibv_flow *flow)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int attach_mcast(struct ibv_qp *qp, const union ibv_gid *gid,
 			uint16_t lid)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int bind_mw(struct ibv_qp *qp, struct ibv_mw *mw,
 		   struct ibv_mw_bind *mw_bind)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int close_xrcd(struct ibv_xrcd *xrcd)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static void cq_event(struct ibv_cq *cq)
@@ -117,14 +117,14 @@ static void cq_event(struct ibv_cq *cq)
 
 static struct ibv_ah *create_ah(struct ibv_pd *pd, struct ibv_ah_attr *attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_counters *create_counters(struct ibv_context *context,
 					    struct ibv_counters_init_attr *init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
@@ -132,42 +132,42 @@ static struct ibv_cq *create_cq(struct ibv_context *context, int cqe,
 				struct ibv_comp_channel *channel,
 				int comp_vector)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_cq_ex *create_cq_ex(struct ibv_context *context,
 				      struct ibv_cq_init_attr_ex *init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_flow *create_flow(struct ibv_qp *qp,
 				    struct ibv_flow_attr *flow_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_flow_action *create_flow_action_esp(struct ibv_context *context,
 						      struct ibv_flow_action_esp_attr *attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_qp *create_qp(struct ibv_pd *pd,
 				struct ibv_qp_init_attr *attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_qp *create_qp_ex(struct ibv_context *context,
 				   struct ibv_qp_init_attr_ex *qp_init_attr_ex)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
@@ -175,14 +175,14 @@ static struct ibv_rwq_ind_table *
 create_rwq_ind_table(struct ibv_context *context,
 		     struct ibv_rwq_ind_table_init_attr *init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_srq *create_srq(struct ibv_pd *pd,
 				  struct ibv_srq_init_attr *srq_init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
@@ -190,208 +190,213 @@ static struct ibv_srq *
 create_srq_ex(struct ibv_context *context,
 	      struct ibv_srq_init_attr_ex *srq_init_attr_ex)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_wq *create_wq(struct ibv_context *context,
 				struct ibv_wq_init_attr *wq_init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static int dealloc_mw(struct ibv_mw *mw)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int dealloc_pd(struct ibv_pd *pd)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int dealloc_td(struct ibv_td *td)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int dereg_mr(struct verbs_mr *vmr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_ah(struct ibv_ah *ah)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_counters(struct ibv_counters *counters)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_cq(struct ibv_cq *cq)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_flow(struct ibv_flow *flow)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_flow_action(struct ibv_flow_action *action)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_qp(struct ibv_qp *qp)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_rwq_ind_table(struct ibv_rwq_ind_table *rwq_ind_table)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_srq(struct ibv_srq *srq)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int destroy_wq(struct ibv_wq *wq)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int detach_mcast(struct ibv_qp *qp, const union ibv_gid *gid,
 			uint16_t lid)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
+}
+
+static void free_context(struct ibv_context *ctx)
+{
+	return;
 }
 
 static int free_dm(struct ibv_dm *dm)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int get_srq_num(struct ibv_srq *srq, uint32_t *srq_num)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int modify_cq(struct ibv_cq *cq, struct ibv_modify_cq_attr *attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int modify_flow_action_esp(struct ibv_flow_action *action,
 				  struct ibv_flow_action_esp_attr *attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr, int attr_mask)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int modify_qp_rate_limit(struct ibv_qp *qp,
 				struct ibv_qp_rate_limit_attr *attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int modify_srq(struct ibv_srq *srq, struct ibv_srq_attr *srq_attr,
 		      int srq_attr_mask)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int modify_wq(struct ibv_wq *wq, struct ibv_wq_attr *wq_attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static struct ibv_qp *open_qp(struct ibv_context *context,
 			      struct ibv_qp_open_attr *attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_xrcd *open_xrcd(struct ibv_context *context,
 				  struct ibv_xrcd_init_attr *xrcd_init_attr)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static int poll_cq(struct ibv_cq *cq, int num_entries, struct ibv_wc *wc)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int post_recv(struct ibv_qp *qp, struct ibv_recv_wr *wr,
 		     struct ibv_recv_wr **bad_wr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int post_send(struct ibv_qp *qp, struct ibv_send_wr *wr,
 		     struct ibv_send_wr **bad_wr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int post_srq_ops(struct ibv_srq *srq, struct ibv_ops_wr *op,
 			struct ibv_ops_wr **bad_op)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int post_srq_recv(struct ibv_srq *srq, struct ibv_recv_wr *recv_wr,
 			 struct ibv_recv_wr **bad_recv_wr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int query_device(struct ibv_context *context,
 			struct ibv_device_attr *device_attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int query_device_ex(struct ibv_context *context,
 			   const struct ibv_query_device_ex_input *input,
 			   struct ibv_device_attr_ex *attr, size_t attr_size)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int query_port(struct ibv_context *context, uint8_t port_num,
 		      struct ibv_port_attr *port_attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int query_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr, int attr_mask,
 		    struct ibv_qp_init_attr *init_attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int query_rt_values(struct ibv_context *context,
 			   struct ibv_values_ex *values)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int query_srq(struct ibv_srq *srq, struct ibv_srq_attr *srq_attr)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int read_counters(struct ibv_counters *counters,
@@ -399,43 +404,43 @@ static int read_counters(struct ibv_counters *counters,
 			 uint32_t ncounters,
 			 uint32_t flags)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static struct ibv_mr *reg_dm_mr(struct ibv_pd *pd, struct ibv_dm *dm,
 				uint64_t dm_offset, size_t length,
 				unsigned int access)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static struct ibv_mr *reg_mr(struct ibv_pd *pd, void *addr, size_t length,
-			     int access)
+			     uint64_t hca_va,  int access)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return NULL;
 }
 
 static int req_notify_cq(struct ibv_cq *cq, int solicited_only)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 static int rereg_mr(struct verbs_mr *vmr, int flags, struct ibv_pd *pd,
 		    void *addr, size_t length, int access)
 {
-	errno = ENOSYS;
+	errno = EOPNOTSUPP;
 	return IBV_REREG_MR_ERR_INPUT;
 }
 
 static int resize_cq(struct ibv_cq *cq, int cqe)
 {
-	return ENOSYS;
+	return EOPNOTSUPP;
 }
 
 /*
- * Ops in verbs_dummy_ops simply return an ENOSYS error code when called, or
+ * Ops in verbs_dummy_ops simply return an EOPNOTSUPP error code when called, or
  * do nothing. They are placed in the ops structures if the provider does not
  * provide an op for the function.
  *
@@ -485,6 +490,7 @@ const struct verbs_context_ops verbs_dummy_ops = {
 	destroy_srq,
 	destroy_wq,
 	detach_mcast,
+	free_context,
 	free_dm,
 	get_srq_num,
 	modify_cq,
@@ -600,6 +606,7 @@ void verbs_set_ops(struct verbs_context *vctx,
 	SET_PRIV_OP(ctx, destroy_srq);
 	SET_OP(vctx, destroy_wq);
 	SET_PRIV_OP(ctx, detach_mcast);
+	SET_PRIV_OP_IC(ctx, free_context);
 	SET_OP(vctx, free_dm);
 	SET_OP(vctx, get_srq_num);
 	SET_OP(vctx, modify_cq);
