@@ -49,11 +49,9 @@
 #include <infiniband/cmd_write.h>
 
 #include "ibverbs.h"
-#ifndef NRESOLVE_NEIGH
 #include <net/if.h>
 #include <net/if_arp.h>
 #include "neigh.h"
-#endif
 
 #undef ibv_query_port
 
@@ -223,15 +221,13 @@ LATEST_SYMVER_FUNC(ibv_query_gid, 1_1, "IBVERBS_1.1",
 		   struct ibv_context *context, uint8_t port_num,
 		   int index, union ibv_gid *gid)
 {
-	char name[24];
+	struct verbs_device *verbs_device = verbs_get_device(context->device);
 	char attr[41];
 	uint16_t val;
 	int i;
 
-	snprintf(name, sizeof name, "ports/%d/gids/%d", port_num, index);
-
-	if (ibv_read_sysfs_file(context->device->ibdev_path, name,
-				attr, sizeof attr) < 0)
+	if (ibv_read_ibdev_sysfs_file(attr, sizeof(attr), verbs_device->sysfs,
+				      "ports/%d/gids/%d", port_num, index) < 0)
 		return -1;
 
 	for (i = 0; i < 8; ++i) {
@@ -249,14 +245,12 @@ LATEST_SYMVER_FUNC(ibv_query_pkey, 1_1, "IBVERBS_1.1",
 		   struct ibv_context *context, uint8_t port_num,
 		   int index, __be16 *pkey)
 {
-	char name[24];
+	struct verbs_device *verbs_device = verbs_get_device(context->device);
 	char attr[8];
 	uint16_t val;
 
-	snprintf(name, sizeof name, "ports/%d/pkeys/%d", port_num, index);
-
-	if (ibv_read_sysfs_file(context->device->ibdev_path, name,
-				attr, sizeof attr) < 0)
+	if (ibv_read_ibdev_sysfs_file(attr, sizeof(attr), verbs_device->sysfs,
+				      "ports/%d/pkeys/%d", port_num, index) < 0)
 		return -1;
 
 	if (sscanf(attr, "%hx", &val) != 1)
@@ -302,6 +296,7 @@ LATEST_SYMVER_FUNC(ibv_dealloc_pd, 1_1, "IBVERBS_1.1",
 	return get_ops(pd->context)->dealloc_pd(pd);
 }
 
+#undef ibv_reg_mr
 LATEST_SYMVER_FUNC(ibv_reg_mr, 1_1, "IBVERBS_1.1",
 		   struct ibv_mr *,
 		   struct ibv_pd *pd, void *addr,
@@ -312,7 +307,8 @@ LATEST_SYMVER_FUNC(ibv_reg_mr, 1_1, "IBVERBS_1.1",
 	if (ibv_dontfork_range(addr, length))
 		return NULL;
 
-	mr = get_ops(pd->context)->reg_mr(pd, addr, length, access);
+	mr = get_ops(pd->context)->reg_mr(pd, addr, length, (uintptr_t) addr,
+					  access);
 	if (mr) {
 		mr->context = pd->context;
 		mr->pd      = pd;
@@ -324,6 +320,38 @@ LATEST_SYMVER_FUNC(ibv_reg_mr, 1_1, "IBVERBS_1.1",
 	return mr;
 }
 
+#undef ibv_reg_mr_iova
+struct ibv_mr *ibv_reg_mr_iova(struct ibv_pd *pd, void *addr, size_t length,
+			       uint64_t iova, int access)
+{
+	struct ibv_mr *mr;
+
+	if (ibv_dontfork_range(addr, length))
+		return NULL;
+
+	mr = get_ops(pd->context)->reg_mr(pd, addr, length, iova, access);
+	if (mr) {
+		mr->context = pd->context;
+		mr->pd      = pd;
+		mr->addr    = addr;
+		mr->length  = length;
+	} else
+		ibv_dofork_range(addr, length);
+
+	return mr;
+}
+
+struct ibv_mr *ibv_reg_mr_iova2(struct ibv_pd *pd, void *addr, size_t length,
+				uint64_t iova, unsigned int access)
+{
+	struct verbs_device *device = verbs_get_device(pd->context->device);
+
+	if (!(device->core_support & IB_UVERBS_CORE_SUPPORT_OPTIONAL_MR_ACCESS))
+		access &= ~IBV_ACCESS_OPTIONAL_RANGE;
+
+	return ibv_reg_mr_iova(pd, addr, length, iova, access);
+}
+
 LATEST_SYMVER_FUNC(ibv_rereg_mr, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_mr *mr, int flags,
@@ -662,18 +690,16 @@ LATEST_SYMVER_FUNC(ibv_create_ah, 1_1, "IBVERBS_1.1",
 int ibv_query_gid_type(struct ibv_context *context, uint8_t port_num,
 		       unsigned int index, enum ibv_gid_type *type)
 {
-	char name[32];
+	struct verbs_device *verbs_device = verbs_get_device(context->device);
 	char buff[11];
 
-	snprintf(name, sizeof(name), "ports/%d/gid_attrs/types/%d", port_num,
-		 index);
-
 	/* Reset errno so that we can rely on its value upon any error flow in
 	 * ibv_read_sysfs_file.
 	 */
 	errno = 0;
-	if (ibv_read_sysfs_file(context->device->ibdev_path, name, buff,
-				sizeof(buff)) <= 0) {
+	if (ibv_read_ibdev_sysfs_file(buff, sizeof(buff), verbs_device->sysfs,
+				      "ports/%d/gid_attrs/types/%d", port_num,
+				      index) <= 0) {
 		char *dir_path;
 		DIR *dir;
 
@@ -685,7 +711,7 @@ int ibv_query_gid_type(struct ibv_context *context, uint8_t port_num,
 			return 0;
 		}
 		if (asprintf(&dir_path, "%s/%s/%d/%s/",
-			     context->device->ibdev_path, "ports", port_num,
+			     verbs_device->sysfs->ibdev_path, "ports", port_num,
 			     "gid_attrs") < 0)
 			return -1;
 		dir = opendir(dir_path);
@@ -973,7 +999,6 @@ int ibv_resolve_eth_l2_from_gid(struct ibv_context *context,
 				uint8_t eth_mac[ETHERNET_LL_SIZE],
 				uint16_t *vid)
 {
-#ifndef NRESOLVE_NEIGH
 	int dst_family;
 	int src_family;
 	int oif;
@@ -1054,7 +1079,4 @@ free_resources:
 	neigh_free_resources(&neigh_handler);
 
 	return ret;
-#else
-	return -ENOSYS;
-#endif
 }
