@@ -8,10 +8,13 @@
 #include <string.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <util/util.h>
 
 #include "efa.h"
 #include "verbs.h"
 
+static void efa_free_context(struct ibv_context *ibvctx);
+
 #define PCI_VENDOR_ID_AMAZON 0x1d0f
 
 static const struct verbs_match_ent efa_table[] = {
@@ -25,6 +28,7 @@ static const struct verbs_context_ops efa_ctx_ops = {
 	.create_ah = efa_create_ah,
 	.create_cq = efa_create_cq,
 	.create_qp = efa_create_qp,
+	.create_qp_ex = efa_create_qp_ex,
 	.dealloc_pd = efa_dealloc_pd,
 	.dereg_mr = efa_dereg_mr,
 	.destroy_ah = efa_destroy_ah,
@@ -39,15 +43,17 @@ static const struct verbs_context_ops efa_ctx_ops = {
 	.query_port = efa_query_port,
 	.query_qp = efa_query_qp,
 	.reg_mr = efa_reg_mr,
+	.free_context = efa_free_context,
 };
 
 static struct verbs_context *efa_alloc_context(struct ibv_device *vdev,
 					       int cmd_fd,
 					       void *private_data)
 {
-	struct efa_alloc_ucontext_resp resp;
+	struct efa_alloc_ucontext_resp resp = {};
 	struct ibv_device_attr_ex attr;
 	struct ibv_get_context cmd;
+	unsigned int qp_table_sz;
 	struct efa_context *ctx;
 	int err;
 
@@ -56,10 +62,9 @@ static struct verbs_context *efa_alloc_context(struct ibv_device *vdev,
 	if (!ctx)
 		return NULL;
 
-	memset(&resp, 0, sizeof(resp));
 	if (ibv_cmd_get_context(&ctx->ibvctx, &cmd, sizeof(cmd),
 				&resp.ibv_resp, sizeof(resp)))
-		goto failed;
+		goto err_free_ctx;
 
 	ctx->sub_cqs_per_cq = resp.sub_cqs_per_cq;
 	ctx->cmds_supp_udata_mask = resp.cmds_supp_udata_mask;
@@ -68,21 +73,28 @@ static struct verbs_context *efa_alloc_context(struct ibv_device *vdev,
 	ctx->max_llq_size = resp.max_llq_size;
 	pthread_spin_init(&ctx->qp_table_lock, PTHREAD_PROCESS_PRIVATE);
 
+	/* ah udata is mandatory for ah number retrieval */
+	if (!(ctx->cmds_supp_udata_mask & EFA_USER_CMDS_SUPP_UDATA_CREATE_AH))
+		goto err_free_spinlock;
+
 	verbs_set_ops(&ctx->ibvctx, &efa_ctx_ops);
 
 	err = efa_query_device_ex(&ctx->ibvctx.context, NULL, &attr,
 				  sizeof(attr));
 	if (err)
-		goto failed;
+		goto err_free_spinlock;
 
-	ctx->qp_table = calloc(attr.orig_attr.max_qp, sizeof(*ctx->qp_table));
+	qp_table_sz = roundup_pow_of_two(attr.orig_attr.max_qp);
+	ctx->qp_table_sz_m1 = qp_table_sz - 1;
+	ctx->qp_table = calloc(qp_table_sz, sizeof(*ctx->qp_table));
 	if (!ctx->qp_table)
-		goto failed;
+		goto err_free_spinlock;
 
 	return &ctx->ibvctx;
 
-failed:
+err_free_spinlock:
 	pthread_spin_destroy(&ctx->qp_table_lock);
+err_free_ctx:
 	verbs_uninit_context(&ctx->ibvctx);
 	free(ctx);
 	return NULL;
@@ -92,6 +104,7 @@ static void efa_free_context(struct ibv_context *ibvctx)
 {
 	struct efa_context *ctx = to_efa_context(ibvctx);
 
+	free(ctx->qp_table);
 	pthread_spin_destroy(&ctx->qp_table_lock);
 	verbs_uninit_context(&ctx->ibvctx);
 	free(ctx);
@@ -125,7 +138,6 @@ static const struct verbs_device_ops efa_dev_ops = {
 	.alloc_device = efa_device_alloc,
 	.uninit_device = efa_uninit_device,
 	.alloc_context = efa_alloc_context,
-	.free_context = efa_free_context,
 };
 
 bool is_efa_dev(struct ibv_device *device)
