@@ -22,6 +22,23 @@
 #include "efadv.h"
 #include "verbs.h"
 
+static bool is_buf_cleared(void *buf, size_t len)
+{
+	int i;
+
+	for (i = 0; i < len; i++) {
+		if (((uint8_t *)buf)[i])
+			return false;
+	}
+
+	return true;
+}
+
+#define is_ext_cleared(ptr, inlen) \
+	is_buf_cleared(ptr + sizeof(*ptr), inlen - sizeof(*ptr))
+
+#define is_reserved_cleared(reserved) is_buf_cleared(reserved, sizeof(reserved))
+
 int efa_query_device(struct ibv_context *ibvctx,
 		     struct ibv_device_attr *dev_attr)
 {
@@ -48,7 +65,6 @@ int efa_query_port(struct ibv_context *ibvctx, uint8_t port,
 {
 	struct ibv_query_port cmd;
 
-	memset(port_attr, 0, sizeof(struct ibv_port_attr));
 	return ibv_cmd_query_port(ibvctx, port, port_attr, &cmd, sizeof(cmd));
 }
 
@@ -61,9 +77,9 @@ int efa_query_device_ex(struct ibv_context *context,
 	struct efa_dev *dev = to_efa_dev(context->device);
 	int cmd_supp_uhw = ctx->cmds_supp_udata_mask &
 			   EFA_USER_CMDS_SUPP_UDATA_QUERY_DEVICE;
+	struct ibv_device_attr *a = &attr->orig_attr;
 	struct efa_query_device_ex_resp resp = {};
-	struct ibv_query_device_ex cmd;
-	struct ibv_device_attr *a;
+	struct ibv_query_device_ex cmd = {};
 	uint8_t fw_ver[8];
 	int err;
 
@@ -74,12 +90,14 @@ int efa_query_device_ex(struct ibv_context *context,
 	if (err)
 		return err;
 
+	dev->device_caps = resp.device_caps;
 	dev->max_sq_wr = resp.max_sq_wr;
 	dev->max_rq_wr = resp.max_rq_wr;
 	dev->max_sq_sge = resp.max_sq_sge;
 	dev->max_rq_sge = resp.max_rq_sge;
+	dev->max_rdma_size = resp.max_rdma_size;
+	dev->max_wr_rdma_sge = a->max_sge_rd;
 
-	a = &attr->orig_attr;
 	a->max_qp_wr = min_t(int, a->max_qp_wr,
 			     ctx->max_llq_size / sizeof(struct efa_io_tx_wqe));
 	snprintf(a->fw_ver, sizeof(a->fw_ver), "%u.%u.%u.%u",
@@ -88,27 +106,62 @@ int efa_query_device_ex(struct ibv_context *context,
 	return 0;
 }
 
+int efadv_query_device(struct ibv_context *ibvctx,
+		       struct efadv_device_attr *attr,
+		       uint32_t inlen)
+{
+	struct efa_context *ctx = to_efa_context(ibvctx);
+	struct efa_dev *dev = to_efa_dev(ibvctx->device);
+	uint64_t comp_mask_out = 0;
+
+	if (!is_efa_dev(ibvctx->device))
+		return EOPNOTSUPP;
+
+	if (!vext_field_avail(typeof(*attr), inline_buf_size, inlen))
+		return EINVAL;
+
+	memset(attr, 0, inlen);
+	attr->max_sq_wr = dev->max_sq_wr;
+	attr->max_rq_wr = dev->max_rq_wr;
+	attr->max_sq_sge = dev->max_sq_sge;
+	attr->max_rq_sge = dev->max_rq_sge;
+	attr->inline_buf_size = ctx->inline_buf_size;
+
+	if (vext_field_avail(typeof(*attr), max_rdma_size, inlen)) {
+		attr->max_rdma_size = dev->max_rdma_size;
+
+		if (is_rdma_read_cap(dev))
+			attr->device_caps |= EFADV_DEVICE_ATTR_CAPS_RDMA_READ;
+	}
+
+	attr->comp_mask = comp_mask_out;
+
+	return 0;
+}
+
 struct ibv_pd *efa_alloc_pd(struct ibv_context *ibvctx)
 {
 	struct efa_alloc_pd_resp resp = {};
 	struct ibv_alloc_pd cmd;
 	struct efa_pd *pd;
+	int err;
 
 	pd = calloc(1, sizeof(*pd));
 	if (!pd)
 		return NULL;
 
-	if (ibv_cmd_alloc_pd(ibvctx, &pd->ibvpd, &cmd, sizeof(cmd),
-			     &resp.ibv_resp, sizeof(resp)))
+	err = ibv_cmd_alloc_pd(ibvctx, &pd->ibvpd, &cmd, sizeof(cmd),
+			       &resp.ibv_resp, sizeof(resp));
+	if (err)
 		goto out;
 
-	pd->context = to_efa_context(ibvctx);
 	pd->pdn = resp.pdn;
 
 	return &pd->ibvpd;
 
 out:
 	free(pd);
+	errno = err;
 	return NULL;
 }
 
@@ -131,14 +184,17 @@ struct ibv_mr *efa_reg_mr(struct ibv_pd *ibvpd, void *sva, size_t len,
 	struct ib_uverbs_reg_mr_resp resp;
 	struct ibv_reg_mr cmd;
 	struct efa_mr *mr;
+	int err;
 
 	mr = calloc(1, sizeof(*mr));
 	if (!mr)
 		return NULL;
 
-	if (ibv_cmd_reg_mr(ibvpd, sva, len, hca_va, access, &mr->vmr,
-			   &cmd, sizeof(cmd), &resp, sizeof(resp))) {
+	err = ibv_cmd_reg_mr(ibvpd, sva, len, hca_va, access, &mr->vmr,
+			     &cmd, sizeof(cmd), &resp, sizeof(resp));
+	if (err) {
 		free(mr);
+		errno = err;
 		return NULL;
 	}
 
@@ -166,7 +222,7 @@ static uint32_t efa_sub_cq_get_current_index(struct efa_sub_cq *sub_cq)
 static int efa_cqe_is_pending(struct efa_io_cdesc_common *cqe_common,
 				 int phase)
 {
-	return (cqe_common->flags & EFA_IO_CDESC_COMMON_PHASE_MASK) == phase;
+	return EFA_GET(&cqe_common->flags, EFA_IO_CDESC_COMMON_PHASE) == phase;
 }
 
 static struct efa_io_cdesc_common *
@@ -198,6 +254,7 @@ struct ibv_cq *efa_create_cq(struct ibv_context *ibvctx, int ncqe,
 	int sub_buf_size;
 	int sub_cq_size;
 	uint8_t *buf;
+	int err;
 	int i;
 
 	cq = calloc(1, sizeof(*cq) +
@@ -210,10 +267,13 @@ struct ibv_cq *efa_create_cq(struct ibv_context *ibvctx, int ncqe,
 	cmd.cq_entry_size = ctx->cqe_size;
 
 	ncqe = roundup_pow_of_two(ncqe);
-	if (ibv_cmd_create_cq(ibvctx, ncqe, channel, vec,
-			      &cq->ibvcq, &cmd.ibv_cmd, sizeof(cmd),
-			      &resp.ibv_resp, sizeof(resp)))
+	err = ibv_cmd_create_cq(ibvctx, ncqe, channel, vec,
+				&cq->ibvcq, &cmd.ibv_cmd, sizeof(cmd),
+				&resp.ibv_resp, sizeof(resp));
+	if (err) {
+		errno = err;
 		goto err_free_cq;
+	}
 
 	sub_cq_size = cq->ibvcq.cqe;
 	cq->cqn = resp.cq_idx;
@@ -329,12 +389,12 @@ static int efa_poll_sub_cq(struct efa_cq *cq, struct efa_sub_cq *sub_cq,
 		return ENOMEM;
 
 	qpn = cqe->qp_num;
-	if (!*cur_qp || qpn != (*cur_qp)->ibvqp.qp_num) {
+	if (!*cur_qp || qpn != (*cur_qp)->verbs_qp.qp.qp_num) {
 		/* We do not have to take the QP table lock here,
 		 * because CQs will be locked while QPs are removed
 		 * from the table.
 		 */
-		*cur_qp = ctx->qp_table[qpn];
+		*cur_qp = ctx->qp_table[qpn & ctx->qp_table_sz_m1];
 		if (!*cur_qp)
 			return EINVAL;
 	}
@@ -342,7 +402,11 @@ static int efa_poll_sub_cq(struct efa_cq *cq, struct efa_sub_cq *sub_cq,
 	wrid_idx = cqe->req_id;
 	wc->status = to_ibv_status(cqe->status);
 	wc->vendor_err = cqe->status;
-	if (get_efa_io_cdesc_common_q_type(cqe) == EFA_IO_SEND_QUEUE) {
+	wc->wc_flags = 0;
+	wc->qp_num = qpn;
+
+	if (EFA_GET(&cqe->flags, EFA_IO_CDESC_COMMON_Q_TYPE) ==
+	    EFA_IO_SEND_QUEUE) {
 		wq = &(*cur_qp)->sq.wq;
 		wc->opcode = IBV_WC_SEND;
 	} else {
@@ -355,15 +419,20 @@ static int efa_poll_sub_cq(struct efa_cq *cq, struct efa_sub_cq *sub_cq,
 		wc->opcode = IBV_WC_RECV;
 		wc->src_qp = rcqe->src_qp_num;
 		wc->sl = 0;
-		wc->slid = 0;
+		wc->slid = rcqe->ah;
+
+		if (EFA_GET(&cqe->flags, EFA_IO_CDESC_COMMON_HAS_IMM)) {
+			wc->imm_data = htobe32(rcqe->imm);
+			wc->wc_flags |= IBV_WC_WITH_IMM;
+		}
 	}
 
-	wc->wc_flags = 0;
-	wc->qp_num = qpn;
+	pthread_spin_lock(&wq->wqlock);
 	wq->wrid_idx_pool_next--;
 	wq->wrid_idx_pool[wq->wrid_idx_pool_next] = wrid_idx;
 	wc->wr_id = wq->wrid[wrid_idx];
 	wq->wqe_completed++;
+	pthread_spin_unlock(&wq->wqlock);
 
 	return 0;
 }
@@ -394,7 +463,7 @@ static int efa_poll_sub_cqs(struct efa_cq *cq, struct ibv_wc *wc)
 int efa_poll_cq(struct ibv_cq *ibvcq, int nwc, struct ibv_wc *wc)
 {
 	struct efa_cq *cq = to_efa_cq(ibvcq);
-	ssize_t ret = 0;
+	int ret = 0;
 	int i;
 
 	pthread_spin_lock(&cq->lock);
@@ -467,12 +536,14 @@ static void efa_sq_terminate(struct efa_qp *qp)
 	db_aligned = (void *)((uintptr_t)qp->sq.db & ~(qp->page_size - 1));
 	munmap(db_aligned, qp->page_size);
 	munmap(qp->sq.desc - qp->sq.desc_offset, qp->sq.desc_ring_mmap_size);
+	free(qp->sq.local_queue);
 
 	efa_wq_terminate(&qp->sq.wq);
 }
 
 static int efa_sq_initialize(struct efa_qp *qp, struct efa_create_qp_resp *resp)
 {
+	struct efa_dev *dev = to_efa_dev(qp->verbs_qp.qp.context->device);
 	size_t desc_ring_size;
 	uint8_t *db_base;
 	int err;
@@ -490,29 +561,43 @@ static int efa_sq_initialize(struct efa_qp *qp, struct efa_create_qp_resp *resp)
 					   qp->page_size);
 	qp->sq.max_inline_data = resp->ibv_resp.max_inline_data;
 
-	qp->sq.desc = mmap(NULL, qp->sq.desc_ring_mmap_size, PROT_WRITE,
-			   MAP_SHARED, qp->ibvqp.context->cmd_fd,
-			   resp->llq_desc_mmap_key);
-	if (qp->sq.desc == MAP_FAILED)
+	qp->sq.local_queue = malloc(desc_ring_size);
+	if (!qp->sq.local_queue) {
+		err = ENOMEM;
 		goto err_terminate_wq;
+	}
+
+	qp->sq.desc = mmap(NULL, qp->sq.desc_ring_mmap_size, PROT_WRITE,
+			   MAP_SHARED, qp->verbs_qp.qp.context->cmd_fd,
+			   resp->llq_desc_mmap_key);
+	if (qp->sq.desc == MAP_FAILED) {
+		err = errno;
+		goto err_free_local_queue;
+	}
 
 	qp->sq.desc += qp->sq.desc_offset;
 
 	db_base = mmap(NULL, qp->page_size, PROT_WRITE, MAP_SHARED,
-		       qp->ibvqp.context->cmd_fd, resp->sq_db_mmap_key);
-	if (db_base == MAP_FAILED)
+		       qp->verbs_qp.qp.context->cmd_fd, resp->sq_db_mmap_key);
+	if (db_base == MAP_FAILED) {
+		err = errno;
 		goto err_unmap_desc_ring;
+	}
 
 	qp->sq.db = (uint32_t *)(db_base + resp->sq_db_offset);
 	qp->sq.sub_cq_idx = resp->send_sub_cq_idx;
+	qp->sq.max_wr_rdma_sge = min_t(uint16_t, dev->max_wr_rdma_sge,
+				       EFA_IO_TX_DESC_NUM_RDMA_BUFS);
 
 	return 0;
 
 err_unmap_desc_ring:
 	munmap(qp->sq.desc - qp->sq.desc_offset, qp->sq.desc_ring_mmap_size);
+err_free_local_queue:
+	free(qp->sq.local_queue);
 err_terminate_wq:
 	efa_wq_terminate(&qp->sq.wq);
-	return EINVAL;
+	return err;
 }
 
 static void efa_rq_terminate(struct efa_qp *qp)
@@ -543,14 +628,18 @@ static int efa_rq_initialize(struct efa_qp *qp, struct efa_create_qp_resp *resp)
 
 	qp->rq.buf_size = resp->rq_mmap_size;
 	qp->rq.buf = mmap(NULL, qp->rq.buf_size, PROT_WRITE, MAP_SHARED,
-			  qp->ibvqp.context->cmd_fd, resp->rq_mmap_key);
-	if (qp->rq.buf == MAP_FAILED)
+			  qp->verbs_qp.qp.context->cmd_fd, resp->rq_mmap_key);
+	if (qp->rq.buf == MAP_FAILED) {
+		err = errno;
 		goto err_terminate_wq;
+	}
 
 	db_base = mmap(NULL, qp->page_size, PROT_WRITE, MAP_SHARED,
-		       qp->ibvqp.context->cmd_fd, resp->rq_db_mmap_key);
-	if (db_base == MAP_FAILED)
+		       qp->verbs_qp.qp.context->cmd_fd, resp->rq_db_mmap_key);
+	if (db_base == MAP_FAILED) {
+		err = errno;
 		goto err_unmap_rq_buf;
+	}
 
 	qp->rq.db = (uint32_t *)(db_base + resp->rq_db_offset);
 	qp->rq.sub_cq_idx = resp->recv_sub_cq_idx;
@@ -561,7 +650,7 @@ err_unmap_rq_buf:
 	munmap(qp->rq.buf, qp->rq.buf_size);
 err_terminate_wq:
 	efa_wq_terminate(&qp->rq.wq);
-	return EINVAL;
+	return err;
 }
 
 static void efa_qp_init_indices(struct efa_qp *qp)
@@ -627,9 +716,49 @@ static void efa_unlock_cqs(struct ibv_qp *ibvqp)
 	}
 }
 
+static void efa_qp_fill_wr_pfns(struct ibv_qp_ex *ibvqpx,
+				struct ibv_qp_init_attr_ex *attr_ex);
+
 static int efa_check_qp_attr(struct efa_dev *dev,
-			     struct ibv_qp_init_attr *attr)
+			     struct ibv_qp_init_attr_ex *attr,
+			     struct efadv_qp_init_attr *efa_attr)
 {
+	uint64_t supp_send_ops_mask;
+	uint64_t supp_ud_send_ops_mask = IBV_QP_EX_WITH_SEND |
+		IBV_QP_EX_WITH_SEND_WITH_IMM;
+	uint64_t supp_srd_send_ops_mask =
+		IBV_QP_EX_WITH_SEND | IBV_QP_EX_WITH_SEND_WITH_IMM |
+		(is_rdma_read_cap(dev) ? IBV_QP_EX_WITH_RDMA_READ : 0);
+
+#define EFA_CREATE_QP_SUPP_ATTR_MASK \
+	(IBV_QP_INIT_ATTR_PD | IBV_QP_INIT_ATTR_SEND_OPS_FLAGS)
+
+	if (attr->qp_type == IBV_QPT_DRIVER &&
+	    efa_attr->driver_qp_type != EFADV_QP_DRIVER_TYPE_SRD)
+		return EOPNOTSUPP;
+
+	if (!check_comp_mask(attr->comp_mask, EFA_CREATE_QP_SUPP_ATTR_MASK))
+		return EOPNOTSUPP;
+
+	if (!(attr->comp_mask & IBV_QP_INIT_ATTR_PD))
+		return EINVAL;
+
+	if (attr->comp_mask & IBV_QP_INIT_ATTR_SEND_OPS_FLAGS) {
+		switch (attr->qp_type) {
+		case IBV_QPT_UD:
+			supp_send_ops_mask = supp_ud_send_ops_mask;
+			break;
+		case IBV_QPT_DRIVER:
+			supp_send_ops_mask = supp_srd_send_ops_mask;
+			break;
+		default:
+			return EOPNOTSUPP;
+		}
+
+		if (!check_comp_mask(attr->send_ops_flags, supp_send_ops_mask))
+			return EOPNOTSUPP;
+	}
+
 	if (!attr->recv_cq || !attr->send_cq)
 		return EINVAL;
 
@@ -640,7 +769,7 @@ static int efa_check_qp_attr(struct efa_dev *dev,
 }
 
 static int efa_check_qp_limits(struct efa_dev *dev,
-			       struct ibv_qp_init_attr *attr)
+			       struct ibv_qp_init_attr_ex *attr)
 {
 	if (attr->cap.max_send_sge > dev->max_sq_sge)
 		return EINVAL;
@@ -657,30 +786,33 @@ static int efa_check_qp_limits(struct efa_dev *dev,
 	return 0;
 }
 
-static struct ibv_qp *create_qp(struct ibv_pd *ibvpd,
-				struct ibv_qp_init_attr *attr,
-				uint32_t driver_qp_type)
+static struct ibv_qp *create_qp(struct ibv_context *ibvctx,
+				struct ibv_qp_init_attr_ex *attr,
+				struct efadv_qp_init_attr *efa_attr)
 {
-	struct efa_context *ctx = to_efa_context(ibvpd->context);
-	struct efa_dev *dev = to_efa_dev(ibvpd->context->device);
+	struct efa_context *ctx = to_efa_context(ibvctx);
+	struct efa_dev *dev = to_efa_dev(ibvctx->device);
 	struct efa_create_qp_resp resp = {};
 	struct efa_create_qp req = {};
 	struct efa_cq *send_cq;
 	struct efa_cq *recv_cq;
+	struct ibv_qp *ibvqp;
 	struct efa_qp *qp;
 	int err;
 
-	err = efa_check_qp_attr(dev, attr);
+	err = efa_check_qp_attr(dev, attr, efa_attr);
 	if (err)
-		return NULL;
+		goto err_out;
 
 	err = efa_check_qp_limits(dev, attr);
 	if (err)
-		return NULL;
+		goto err_out;
 
 	qp = calloc(1, sizeof(*qp));
-	if (!qp)
-		return NULL;
+	if (!qp) {
+		err = ENOMEM;
+		goto err_out;
+	}
 
 	efa_setup_qp(qp, &attr->cap, dev->pg_sz);
 
@@ -692,15 +824,17 @@ static struct ibv_qp *create_qp(struct ibv_pd *ibvpd,
 	req.sq_ring_size = (attr->cap.max_send_wr) *
 		sizeof(struct efa_io_tx_wqe);
 	if (attr->qp_type == IBV_QPT_DRIVER)
-		req.driver_qp_type = driver_qp_type;
+		req.driver_qp_type = efa_attr->driver_qp_type;
 
-	if (ibv_cmd_create_qp(ibvpd, &qp->ibvqp, attr, &req.ibv_cmd,
-			      sizeof(req), &resp.ibv_resp, sizeof(resp)))
+	err = ibv_cmd_create_qp_ex(ibvctx, &qp->verbs_qp, sizeof(qp->verbs_qp),
+				   attr, &req.ibv_cmd, sizeof(req),
+				   &resp.ibv_resp, sizeof(resp));
+	if (err)
 		goto err_free_qp;
 
-	qp->ibvqp.state = IBV_QPS_RESET;
+	ibvqp = &qp->verbs_qp.qp;
+	ibvqp->state = IBV_QPS_RESET;
 	qp->sq_sig_all = attr->sq_sig_all;
-	qp->ctx = ctx;
 
 	err = efa_rq_initialize(qp, &resp);
 	if (err)
@@ -711,7 +845,7 @@ static struct ibv_qp *create_qp(struct ibv_pd *ibvpd,
 		goto err_terminate_rq;
 
 	pthread_spin_lock(&ctx->qp_table_lock);
-	ctx->qp_table[qp->ibvqp.qp_num] = qp;
+	ctx->qp_table[ibvqp->qp_num & ctx->qp_table_sz_m1] = qp;
 	pthread_spin_unlock(&ctx->qp_table_lock);
 
 	if (attr->send_cq) {
@@ -730,42 +864,115 @@ static struct ibv_qp *create_qp(struct ibv_pd *ibvpd,
 		pthread_spin_unlock(&recv_cq->lock);
 	}
 
-	return &qp->ibvqp;
+	if (attr->comp_mask & IBV_QP_INIT_ATTR_SEND_OPS_FLAGS) {
+		efa_qp_fill_wr_pfns(&qp->verbs_qp.qp_ex, attr);
+		qp->verbs_qp.comp_mask |= VERBS_QP_EX;
+	}
+
+	return ibvqp;
 
 err_terminate_rq:
 	efa_rq_terminate(qp);
 err_destroy_qp:
-	ibv_cmd_destroy_qp(&qp->ibvqp);
+	ibv_cmd_destroy_qp(ibvqp);
 err_free_qp:
 	free(qp);
+err_out:
+	errno = err;
 	return NULL;
 }
 
 struct ibv_qp *efa_create_qp(struct ibv_pd *ibvpd,
 			     struct ibv_qp_init_attr *attr)
 {
-	if (attr->qp_type != IBV_QPT_UD)
-		return NULL;
+	struct ibv_qp_init_attr_ex attr_ex = {};
+	struct ibv_qp *ibvqp;
 
-	return create_qp(ibvpd, attr, 0);
+	if (attr->qp_type != IBV_QPT_UD) {
+		errno = EOPNOTSUPP;
+		return NULL;
+	}
+
+	memcpy(&attr_ex, attr, sizeof(*attr));
+	attr_ex.comp_mask = IBV_QP_INIT_ATTR_PD;
+	attr_ex.pd = ibvpd;
+
+	ibvqp = create_qp(ibvpd->context, &attr_ex, NULL);
+	if (ibvqp)
+		memcpy(attr, &attr_ex, sizeof(*attr));
+
+	return ibvqp;
+}
+
+struct ibv_qp *efa_create_qp_ex(struct ibv_context *ibvctx,
+				struct ibv_qp_init_attr_ex *attr_ex)
+{
+	if (attr_ex->qp_type != IBV_QPT_UD) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	return create_qp(ibvctx, attr_ex, NULL);
 }
 
 struct ibv_qp *efadv_create_driver_qp(struct ibv_pd *ibvpd,
 				      struct ibv_qp_init_attr *attr,
 				      uint32_t driver_qp_type)
 {
-	if (!is_efa_dev(ibvpd->context->device) ||
-	    attr->qp_type != IBV_QPT_DRIVER)
-		return NULL;
+	struct ibv_qp_init_attr_ex attr_ex = {};
+	struct efadv_qp_init_attr efa_attr = {};
+	struct ibv_qp *ibvqp;
 
-	return create_qp(ibvpd, attr, driver_qp_type);
+	if (!is_efa_dev(ibvpd->context->device)) {
+		errno = EOPNOTSUPP;
+		return NULL;
+	}
+
+	if (attr->qp_type != IBV_QPT_DRIVER) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	memcpy(&attr_ex, attr, sizeof(*attr));
+	attr_ex.comp_mask = IBV_QP_INIT_ATTR_PD;
+	attr_ex.pd = ibvpd;
+	efa_attr.driver_qp_type = driver_qp_type;
+
+	ibvqp = create_qp(ibvpd->context, &attr_ex, &efa_attr);
+	if (ibvqp)
+		memcpy(attr, &attr_ex, sizeof(*attr));
+
+	return ibvqp;
+}
+
+struct ibv_qp *efadv_create_qp_ex(struct ibv_context *ibvctx,
+				  struct ibv_qp_init_attr_ex *attr_ex,
+				  struct efadv_qp_init_attr *efa_attr,
+				  uint32_t inlen)
+{
+	if (!is_efa_dev(ibvctx->device)) {
+		errno = EOPNOTSUPP;
+		return NULL;
+	}
+
+	if (attr_ex->qp_type != IBV_QPT_DRIVER ||
+	    !vext_field_avail(struct efadv_qp_init_attr,
+			      driver_qp_type, inlen) ||
+	    efa_attr->comp_mask ||
+	    !is_reserved_cleared(efa_attr->reserved) ||
+	    (inlen > sizeof(efa_attr) && !is_ext_cleared(efa_attr, inlen))) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	return create_qp(ibvctx, attr_ex, efa_attr);
 }
 
 int efa_modify_qp(struct ibv_qp *ibvqp, struct ibv_qp_attr *attr,
 		  int attr_mask)
 {
 	struct efa_qp *qp = to_efa_qp(ibvqp);
-	struct ibv_modify_qp cmd;
+	struct ibv_modify_qp cmd = {};
 	int err;
 
 	err = ibv_cmd_modify_qp(ibvqp, attr, attr_mask, &cmd, sizeof(cmd));
@@ -773,9 +980,9 @@ int efa_modify_qp(struct ibv_qp *ibvqp, struct ibv_qp_attr *attr,
 		return err;
 
 	if (attr_mask & IBV_QP_STATE) {
-		qp->ibvqp.state = attr->qp_state;
+		qp->verbs_qp.qp.state = attr->qp_state;
 		/* transition to reset */
-		if (qp->ibvqp.state == IBV_QPS_RESET)
+		if (qp->verbs_qp.qp.state == IBV_QPS_RESET)
 			efa_qp_init_indices(qp);
 	}
 
@@ -808,7 +1015,7 @@ int efa_destroy_qp(struct ibv_qp *ibvqp)
 		efa_cq_dec_ref_cnt(to_efa_cq(ibvqp->recv_cq),
 				   qp->rq.sub_cq_idx);
 
-	ctx->qp_table[ibvqp->qp_num] = NULL;
+	ctx->qp_table[ibvqp->qp_num & ctx->qp_table_sz_m1] = NULL;
 
 	efa_unlock_cqs(ibvqp);
 	pthread_spin_unlock(&ctx->qp_table_lock);
@@ -824,33 +1031,31 @@ int efa_destroy_qp(struct ibv_qp *ibvqp)
 	return 0;
 }
 
-static void efa_post_send_sgl(struct ibv_send_wr *wr,
-			      struct efa_io_tx_wqe *tx_wqe,
-			      int *desc_size)
+static void efa_set_tx_buf(struct efa_io_tx_buf_desc *tx_buf,
+			   uint64_t addr, uint32_t lkey,
+			   uint32_t length)
 {
-	struct efa_io_tx_buf_desc *tx_buf;
-	struct ibv_sge *sge;
-	uintptr_t addr;
+	tx_buf->length = length;
+	EFA_SET(&tx_buf->lkey, EFA_IO_TX_BUF_DESC_LKEY, lkey);
+	tx_buf->buf_addr_lo = addr & 0xffffffff;
+	tx_buf->buf_addr_hi = addr >> 32;
+}
+
+static void efa_post_send_sgl(struct efa_io_tx_buf_desc *tx_bufs,
+			      const struct ibv_sge *sg_list,
+			      int num_sge)
+{
+	const struct ibv_sge *sge;
 	size_t i;
 
-	for (i = 0; i < wr->num_sge; i++) {
-		sge = &wr->sg_list[i];
-		tx_buf = &tx_wqe->data.sgl[i];
-		addr = sge->addr;
-
-		/* Set TX buffer desc from SGE */
-		tx_buf->length = sge->length;
-		tx_buf->lkey = sge->lkey;
-		tx_buf->buf_addr_lo = addr & 0xffffffff;
-		tx_buf->buf_addr_hi = (uint64_t)addr >> 32;
+	for (i = 0; i < num_sge; i++) {
+		sge = &sg_list[i];
+		efa_set_tx_buf(&tx_bufs[i], sge->addr, sge->lkey, sge->length);
 	}
-
-	*desc_size += sizeof(*tx_buf) * wr->num_sge;
 }
 
 static void efa_post_send_inline_data(const struct ibv_send_wr *wr,
-				      struct efa_io_tx_wqe *tx_wqe,
-				      int *desc_size)
+				      struct efa_io_tx_wqe *tx_wqe)
 {
 	const struct ibv_sge *sgl = wr->sg_list;
 	uint32_t total_length = 0;
@@ -865,47 +1070,83 @@ static void efa_post_send_inline_data(const struct ibv_send_wr *wr,
 		total_length += length;
 	}
 
-	*desc_size += total_length;
-
-	set_efa_io_tx_meta_desc_inline_msg(&tx_wqe->common, 1);
-	tx_wqe->common.length = total_length;
+	EFA_SET(&tx_wqe->meta.ctrl1, EFA_IO_TX_META_DESC_INLINE_MSG, 1);
+	tx_wqe->meta.length = total_length;
 }
 
-static size_t efa_sge_total_bytes(const struct ibv_send_wr *wr)
+static size_t efa_sge_total_bytes(const struct ibv_sge *sg_list, int num_sge)
 {
 	size_t bytes = 0;
 	size_t i;
 
-	for (i = 0; i < wr->num_sge; i++)
-		bytes += wr->sg_list[i].length;
+	for (i = 0; i < num_sge; i++)
+		bytes += sg_list[i].length;
 
 	return bytes;
 }
 
-static ssize_t efa_post_send_validate(struct efa_qp *qp,
-				      const struct ibv_send_wr *wr)
+static size_t efa_buf_list_total_bytes(const struct ibv_data_buf *buf_list,
+				       size_t num_buf)
 {
-	if (unlikely(qp->ibvqp.state != IBV_QPS_RTS &&
-		     qp->ibvqp.state != IBV_QPS_SQD))
-		return EINVAL;
+	size_t bytes = 0;
+	size_t i;
 
-	if (unlikely(wr->opcode != IBV_WR_SEND))
+	for (i = 0; i < num_buf; i++)
+		bytes += buf_list[i].length;
+
+	return bytes;
+}
+
+static void efa_sq_advance_post_idx(struct efa_qp *qp)
+{
+	qp->sq.wq.wqe_posted++;
+	qp->sq.wq.desc_idx++;
+
+	if (!(qp->sq.wq.desc_idx & qp->sq.wq.desc_mask))
+		qp->sq.wq.phase++;
+}
+
+static uint32_t efa_wq_get_next_wrid_idx(struct efa_wq *wq, uint64_t wr_id)
+{
+	uint32_t wrid_idx;
+
+	/* Get the next wrid to be used from the index pool */
+	wrid_idx = wq->wrid_idx_pool[wq->wrid_idx_pool_next];
+	wq->wrid[wrid_idx] = wr_id;
+
+	/* Will never overlap, as validate function succeeded */
+	wq->wrid_idx_pool_next++;
+	assert(wq->wrid_idx_pool_next <= wq->wqe_cnt);
+
+	return wrid_idx;
+}
+
+static void efa_set_common_ctrl_flags(struct efa_io_tx_meta_desc *desc,
+				      struct efa_qp *qp,
+				      enum efa_io_send_op_type op_type)
+{
+	EFA_SET(&desc->ctrl1, EFA_IO_TX_META_DESC_META_DESC, 1);
+	EFA_SET(&desc->ctrl1, EFA_IO_TX_META_DESC_OP_TYPE, op_type);
+	EFA_SET(&desc->ctrl2, EFA_IO_TX_META_DESC_PHASE, qp->sq.wq.phase);
+	EFA_SET(&desc->ctrl2, EFA_IO_TX_META_DESC_FIRST, 1);
+	EFA_SET(&desc->ctrl2, EFA_IO_TX_META_DESC_LAST, 1);
+	EFA_SET(&desc->ctrl2, EFA_IO_TX_META_DESC_COMP_REQ, 1);
+}
+
+static int efa_post_send_validate(struct efa_qp *qp,
+				  unsigned int wr_flags)
+{
+	if (unlikely(qp->verbs_qp.qp.state != IBV_QPS_RTS &&
+		     qp->verbs_qp.qp.state != IBV_QPS_SQD))
 		return EINVAL;
 
 	if (unlikely(!qp->scq))
 		return EINVAL;
 
-	if (unlikely(wr->num_sge > qp->sq.wq.max_sge))
+	if (unlikely(!(wr_flags & IBV_SEND_SIGNALED) && !qp->sq_sig_all))
 		return EINVAL;
 
-	if (unlikely(!(wr->send_flags & IBV_SEND_SIGNALED) && !qp->sq_sig_all))
-		return EINVAL;
-
-	if (unlikely(wr->send_flags & ~(IBV_SEND_SIGNALED | IBV_SEND_INLINE)))
-		return EINVAL;
-
-	if (unlikely(wr->send_flags & IBV_SEND_INLINE &&
-		     efa_sge_total_bytes(wr) > qp->sq.max_inline_data))
+	if (unlikely(wr_flags & ~(IBV_SEND_SIGNALED | IBV_SEND_INLINE)))
 		return EINVAL;
 
 	if (unlikely(qp->sq.wq.wqe_posted - qp->sq.wq.wqe_completed ==
@@ -915,67 +1156,81 @@ static ssize_t efa_post_send_validate(struct efa_qp *qp,
 	return 0;
 }
 
+static int efa_post_send_validate_wr(struct efa_qp *qp,
+				     const struct ibv_send_wr *wr)
+{
+	int err;
+
+	err = efa_post_send_validate(qp, wr->send_flags);
+	if (unlikely(err))
+		return err;
+
+	if (unlikely(wr->opcode != IBV_WR_SEND &&
+		     wr->opcode != IBV_WR_SEND_WITH_IMM))
+		return EINVAL;
+
+	if (wr->send_flags & IBV_SEND_INLINE) {
+		if (unlikely(efa_sge_total_bytes(wr->sg_list, wr->num_sge) >
+			     qp->sq.max_inline_data))
+			return EINVAL;
+	} else {
+		if (unlikely(wr->num_sge > qp->sq.wq.max_sge))
+			return EINVAL;
+	}
+
+	return 0;
+}
+
 int efa_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 		  struct ibv_send_wr **bad)
 {
 	struct efa_io_tx_meta_desc *meta_desc;
 	struct efa_qp *qp = to_efa_qp(ibvqp);
-	uint32_t sq_desc_offset, wrid_idx;
 	struct efa_io_tx_wqe tx_wqe;
+	uint32_t sq_desc_offset;
 	struct efa_ah *ah;
-	int desc_size;
 	int err = 0;
 
 	pthread_spin_lock(&qp->sq.wq.wqlock);
 	while (wr) {
-		desc_size = sizeof(tx_wqe.common) + sizeof(tx_wqe.u);
-
-		err = efa_post_send_validate(qp, wr);
+		err = efa_post_send_validate_wr(qp, wr);
 		if (err) {
 			*bad = wr;
 			goto ring_db;
 		}
 
 		memset(&tx_wqe, 0, sizeof(tx_wqe));
-		meta_desc = &tx_wqe.common;
+		meta_desc = &tx_wqe.meta;
 		ah = to_efa_ah(wr->wr.ud.ah);
 
-		if (efa_sge_total_bytes(wr) <= qp->ctx->inline_buf_size) {
-			efa_post_send_inline_data(wr, &tx_wqe, &desc_size);
+		if (wr->send_flags & IBV_SEND_INLINE) {
+			efa_post_send_inline_data(wr, &tx_wqe);
 		} else {
 			meta_desc->length = wr->num_sge;
-			efa_post_send_sgl(wr, &tx_wqe, &desc_size);
+			efa_post_send_sgl(tx_wqe.data.sgl, wr->sg_list,
+					  wr->num_sge);
 		}
 
-		/* Get the next wrid to be used from the index pool */
-		wrid_idx = qp->sq.wq.wrid_idx_pool[qp->sq.wq.wrid_idx_pool_next];
-		qp->sq.wq.wrid[wrid_idx] = wr->wr_id;
-		meta_desc->req_id = wrid_idx;
-		qp->sq.wq.wqe_posted++;
-
-		/* Will never overlap, as efa_post_send_validate() succeeded */
-		qp->sq.wq.wrid_idx_pool_next++;
-		assert(qp->sq.wq.wrid_idx_pool_next <= qp->sq.wq.wqe_cnt);
+		if (wr->opcode == IBV_WR_SEND_WITH_IMM) {
+			meta_desc->immediate_data = be32toh(wr->imm_data);
+			EFA_SET(&meta_desc->ctrl1, EFA_IO_TX_META_DESC_HAS_IMM,
+				1);
+		}
 
 		/* Set rest of the descriptor fields */
-		set_efa_io_tx_meta_desc_meta_desc(meta_desc, 1);
-		set_efa_io_tx_meta_desc_phase(meta_desc, qp->sq.wq.phase);
-		set_efa_io_tx_meta_desc_first(meta_desc, 1);
-		set_efa_io_tx_meta_desc_last(meta_desc, 1);
+		efa_set_common_ctrl_flags(meta_desc, qp, EFA_IO_SEND);
+		meta_desc->req_id = efa_wq_get_next_wrid_idx(&qp->sq.wq, wr->wr_id);
 		meta_desc->dest_qp_num = wr->wr.ud.remote_qpn;
-		set_efa_io_tx_meta_desc_comp_req(meta_desc, 1);
 		meta_desc->ah = ah->efa_ah;
-		tx_wqe.u.ud.qkey = wr->wr.ud.remote_qkey;
+		meta_desc->qkey = wr->wr.ud.remote_qkey;
 
 		/* Copy descriptor */
 		sq_desc_offset = (qp->sq.wq.desc_idx & qp->sq.wq.desc_mask) *
 				 sizeof(tx_wqe);
-		memcpy(qp->sq.desc + sq_desc_offset, &tx_wqe, desc_size);
+		memcpy(qp->sq.desc + sq_desc_offset, &tx_wqe, sizeof(tx_wqe));
 
 		/* advance index and change phase */
-		qp->sq.wq.desc_idx++;
-		if (!(qp->sq.wq.desc_idx & qp->sq.wq.desc_mask))
-			qp->sq.wq.phase++;
+		efa_sq_advance_post_idx(qp);
 
 		wr = wr->next;
 	}
@@ -988,10 +1243,304 @@ ring_db:
 	return err;
 }
 
-static ssize_t efa_post_recv_validate(struct efa_qp *qp, struct ibv_recv_wr *wr)
+static int efa_send_wr_common(struct ibv_qp_ex *ibvqpx,
+			      enum efa_io_send_op_type op_type)
 {
-	if (unlikely(qp->ibvqp.state == IBV_QPS_RESET ||
-		     qp->ibvqp.state == IBV_QPS_ERR))
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_tx_meta_desc *meta_desc;
+	int err;
+
+	if (unlikely(qp->wr_session_err))
+		return qp->wr_session_err;
+
+	err = efa_post_send_validate(qp, ibvqpx->wr_flags);
+	if (unlikely(err)) {
+		qp->wr_session_err = err;
+		return err;
+	}
+
+	qp->sq.curr_tx_wqe = (struct efa_io_tx_wqe *)qp->sq.local_queue +
+			     qp->sq.num_wqe_pending;
+	memset(qp->sq.curr_tx_wqe, 0, sizeof(*qp->sq.curr_tx_wqe));
+
+	meta_desc = &qp->sq.curr_tx_wqe->meta;
+	efa_set_common_ctrl_flags(meta_desc, qp, op_type);
+	meta_desc->req_id = efa_wq_get_next_wrid_idx(&qp->sq.wq, ibvqpx->wr_id);
+
+	/* advance index and change phase */
+	efa_sq_advance_post_idx(qp);
+	qp->sq.num_wqe_pending++;
+
+	return 0;
+}
+
+static void efa_send_wr_send(struct ibv_qp_ex *ibvqpx)
+{
+	efa_send_wr_common(ibvqpx, EFA_IO_SEND);
+}
+
+static void efa_send_wr_send_imm(struct ibv_qp_ex *ibvqpx, __be32 imm_data)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_tx_meta_desc *meta_desc;
+	int err;
+
+	err = efa_send_wr_common(ibvqpx, EFA_IO_SEND);
+	if (unlikely(err))
+		return;
+
+	meta_desc = &qp->sq.curr_tx_wqe->meta;
+	meta_desc->immediate_data = be32toh(imm_data);
+	EFA_SET(&meta_desc->ctrl1, EFA_IO_TX_META_DESC_HAS_IMM, 1);
+}
+
+static void efa_send_wr_rdma_read(struct ibv_qp_ex *ibvqpx, uint32_t rkey,
+				  uint64_t remote_addr)
+{
+	struct efa_io_remote_mem_addr *remote_mem;
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_tx_wqe *tx_wqe;
+	int err;
+
+	err = efa_send_wr_common(ibvqpx, EFA_IO_RDMA_READ);
+	if (unlikely(err))
+		return;
+
+	tx_wqe = qp->sq.curr_tx_wqe;
+	remote_mem = &tx_wqe->data.rdma_req.remote_mem;
+	remote_mem->rkey = rkey;
+	remote_mem->buf_addr_lo = remote_addr & 0xFFFFFFFF;
+	remote_mem->buf_addr_hi = remote_addr >> 32;
+}
+
+static void efa_send_wr_set_sge(struct ibv_qp_ex *ibvqpx, uint32_t lkey,
+				uint64_t addr, uint32_t length)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_tx_buf_desc *buf;
+	struct efa_io_tx_wqe *tx_wqe;
+	uint8_t op_type;
+
+	if (unlikely(qp->wr_session_err))
+		return;
+
+	tx_wqe = qp->sq.curr_tx_wqe;
+	tx_wqe->meta.length = 1;
+
+	op_type = EFA_GET(&tx_wqe->meta.ctrl1, EFA_IO_TX_META_DESC_OP_TYPE);
+	switch (op_type) {
+	case EFA_IO_SEND:
+		buf = &tx_wqe->data.sgl[0];
+		break;
+	case EFA_IO_RDMA_READ:
+		tx_wqe->data.rdma_req.remote_mem.length = length;
+		buf = &tx_wqe->data.rdma_req.local_mem[0];
+		break;
+	default:
+		return;
+	}
+
+	efa_set_tx_buf(buf, addr, lkey, length);
+}
+
+static void efa_send_wr_set_sge_list(struct ibv_qp_ex *ibvqpx, size_t num_sge,
+				     const struct ibv_sge *sg_list)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_rdma_req *rdma_req;
+	struct efa_io_tx_wqe *tx_wqe;
+	uint8_t op_type;
+
+	if (unlikely(qp->wr_session_err))
+		return;
+
+	tx_wqe = qp->sq.curr_tx_wqe;
+	op_type = EFA_GET(&tx_wqe->meta.ctrl1, EFA_IO_TX_META_DESC_OP_TYPE);
+	switch (op_type) {
+	case EFA_IO_SEND:
+		if (unlikely(num_sge > qp->sq.wq.max_sge)) {
+			qp->wr_session_err = EINVAL;
+			return;
+		}
+		efa_post_send_sgl(tx_wqe->data.sgl, sg_list, num_sge);
+		break;
+	case EFA_IO_RDMA_READ:
+		if (unlikely(num_sge > qp->sq.max_wr_rdma_sge)) {
+			qp->wr_session_err = EINVAL;
+			return;
+		}
+		rdma_req = &tx_wqe->data.rdma_req;
+		rdma_req->remote_mem.length = efa_sge_total_bytes(sg_list,
+								  num_sge);
+		efa_post_send_sgl(rdma_req->local_mem, sg_list, num_sge);
+		break;
+	default:
+		return;
+	}
+
+	tx_wqe->meta.length = num_sge;
+}
+
+static void efa_send_wr_set_inline_data(struct ibv_qp_ex *ibvqpx, void *addr,
+					size_t length)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_tx_wqe *tx_wqe = qp->sq.curr_tx_wqe;
+
+	if (unlikely(qp->wr_session_err))
+		return;
+
+	if (unlikely(length > qp->sq.max_inline_data)) {
+		qp->wr_session_err = EINVAL;
+		return;
+	}
+
+	EFA_SET(&tx_wqe->meta.ctrl1, EFA_IO_TX_META_DESC_INLINE_MSG, 1);
+	memcpy(tx_wqe->data.inline_data, addr, length);
+	tx_wqe->meta.length = length;
+}
+
+static void
+efa_send_wr_set_inline_data_list(struct ibv_qp_ex *ibvqpx,
+				 size_t num_buf,
+				 const struct ibv_data_buf *buf_list)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_io_tx_wqe *tx_wqe = qp->sq.curr_tx_wqe;
+	uint32_t total_length = 0;
+	uint32_t length;
+	size_t i;
+
+	if (unlikely(qp->wr_session_err))
+		return;
+
+	if (unlikely(efa_buf_list_total_bytes(buf_list, num_buf) >
+		     qp->sq.max_inline_data)) {
+		qp->wr_session_err = EINVAL;
+		return;
+	}
+
+	for (i = 0; i < num_buf; i++) {
+		length = buf_list[i].length;
+
+		memcpy(tx_wqe->data.inline_data + total_length,
+		       buf_list[i].addr, length);
+		total_length += length;
+	}
+
+	EFA_SET(&tx_wqe->meta.ctrl1, EFA_IO_TX_META_DESC_INLINE_MSG, 1);
+	tx_wqe->meta.length = total_length;
+}
+
+static void efa_send_wr_set_addr(struct ibv_qp_ex *ibvqpx,
+				 struct ibv_ah *ibvah,
+				 uint32_t remote_qpn, uint32_t remote_qkey)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	struct efa_ah *ah = to_efa_ah(ibvah);
+	struct efa_io_tx_wqe *tx_wqe = qp->sq.curr_tx_wqe;
+
+	if (unlikely(qp->wr_session_err))
+		return;
+
+	tx_wqe->meta.dest_qp_num = remote_qpn;
+	tx_wqe->meta.ah = ah->efa_ah;
+	tx_wqe->meta.qkey = remote_qkey;
+}
+
+static void efa_send_wr_start(struct ibv_qp_ex *ibvqpx)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+
+	pthread_spin_lock(&qp->sq.wq.wqlock);
+	qp->wr_session_err = 0;
+	qp->sq.num_wqe_pending = 0;
+	qp->sq.phase_rb = qp->sq.wq.phase;
+}
+
+static inline void efa_sq_roll_back(struct efa_qp *qp)
+{
+	qp->sq.wq.wqe_posted -= qp->sq.num_wqe_pending;
+	qp->sq.wq.desc_idx -= qp->sq.num_wqe_pending;
+	qp->sq.wq.wrid_idx_pool_next -= qp->sq.num_wqe_pending;
+	qp->sq.wq.phase = qp->sq.phase_rb;
+}
+
+static int efa_send_wr_complete(struct ibv_qp_ex *ibvqpx)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+	uint32_t num_wqe_to_copy;
+	uint16_t local_idx = 0;
+	uint16_t sq_desc_idx;
+
+	if (unlikely(qp->wr_session_err)) {
+		efa_sq_roll_back(qp);
+		goto out;
+	}
+
+	/*
+	 * Copy local queue to device in chunks, as the descriptor index
+	 * might have wrapped around the submission queue.
+	 */
+	sq_desc_idx = (qp->sq.wq.desc_idx - qp->sq.num_wqe_pending) &
+		       qp->sq.wq.desc_mask;
+
+	while (qp->sq.num_wqe_pending) {
+		num_wqe_to_copy = min(qp->sq.num_wqe_pending,
+				      qp->sq.wq.wqe_cnt - sq_desc_idx);
+		memcpy((struct efa_io_tx_wqe *)qp->sq.desc + sq_desc_idx,
+		       (struct efa_io_tx_wqe *)qp->sq.local_queue + local_idx,
+		       num_wqe_to_copy * sizeof(struct efa_io_tx_wqe));
+
+		qp->sq.num_wqe_pending -= num_wqe_to_copy;
+		local_idx += num_wqe_to_copy;
+		sq_desc_idx = (sq_desc_idx + num_wqe_to_copy) &
+			      qp->sq.wq.desc_mask;
+	}
+
+	udma_to_device_barrier();
+	mmio_write32(qp->sq.db, qp->sq.wq.desc_idx);
+out:
+	pthread_spin_unlock(&qp->sq.wq.wqlock);
+
+	return qp->wr_session_err;
+}
+
+static void efa_send_wr_abort(struct ibv_qp_ex *ibvqpx)
+{
+	struct efa_qp *qp = to_efa_qp_ex(ibvqpx);
+
+	efa_sq_roll_back(qp);
+	pthread_spin_unlock(&qp->sq.wq.wqlock);
+}
+
+static void efa_qp_fill_wr_pfns(struct ibv_qp_ex *ibvqpx,
+				struct ibv_qp_init_attr_ex *attr_ex)
+{
+	ibvqpx->wr_start = efa_send_wr_start;
+	ibvqpx->wr_complete = efa_send_wr_complete;
+	ibvqpx->wr_abort = efa_send_wr_abort;
+
+	if (attr_ex->send_ops_flags & IBV_QP_EX_WITH_SEND)
+		ibvqpx->wr_send = efa_send_wr_send;
+
+	if (attr_ex->send_ops_flags & IBV_QP_EX_WITH_SEND_WITH_IMM)
+		ibvqpx->wr_send_imm = efa_send_wr_send_imm;
+
+	if (attr_ex->send_ops_flags & IBV_QP_EX_WITH_RDMA_READ)
+		ibvqpx->wr_rdma_read = efa_send_wr_rdma_read;
+
+	ibvqpx->wr_set_inline_data = efa_send_wr_set_inline_data;
+	ibvqpx->wr_set_inline_data_list = efa_send_wr_set_inline_data_list;
+	ibvqpx->wr_set_sge = efa_send_wr_set_sge;
+	ibvqpx->wr_set_sge_list = efa_send_wr_set_sge_list;
+	ibvqpx->wr_set_ud_addr = efa_send_wr_set_addr;
+}
+
+static int efa_post_recv_validate(struct efa_qp *qp, struct ibv_recv_wr *wr)
+{
+	if (unlikely(qp->verbs_qp.qp.state == IBV_QPS_RESET ||
+		     qp->verbs_qp.qp.state == IBV_QPS_ERR))
 		return EINVAL;
 
 	if (unlikely(!qp->rcq))
@@ -1011,8 +1560,8 @@ int efa_post_recv(struct ibv_qp *ibvqp, struct ibv_recv_wr *wr,
 		  struct ibv_recv_wr **bad)
 {
 	struct efa_qp *qp = to_efa_qp(ibvqp);
-	uint32_t wqe_index, rq_desc_offset;
 	struct efa_io_rx_desc rx_buf;
+	uint32_t rq_desc_offset;
 	uintptr_t addr;
 	int err = 0;
 	size_t i;
@@ -1027,31 +1576,25 @@ int efa_post_recv(struct ibv_qp *ibvqp, struct ibv_recv_wr *wr,
 
 		memset(&rx_buf, 0, sizeof(rx_buf));
 
-		/* Save wrid */
-		/* Get the next wrid to be used from the index pool */
-		wqe_index = qp->rq.wq.wrid_idx_pool[qp->rq.wq.wrid_idx_pool_next];
-		qp->rq.wq.wrid[wqe_index] = wr->wr_id;
-		rx_buf.req_id = wqe_index;
+		rx_buf.req_id = efa_wq_get_next_wrid_idx(&qp->rq.wq, wr->wr_id);
 		qp->rq.wq.wqe_posted++;
 
-		/* Will never overlap, as efa_post_recv_validate() succeeded */
-		qp->rq.wq.wrid_idx_pool_next++;
-		assert(qp->rq.wq.wrid_idx_pool_next <= qp->rq.wq.wqe_cnt);
-
 		/* Default init of the rx buffer */
-		set_efa_io_rx_desc_first(&rx_buf, 1);
-		set_efa_io_rx_desc_last(&rx_buf, 0);
+		EFA_SET(&rx_buf.lkey_ctrl, EFA_IO_RX_DESC_FIRST, 1);
+		EFA_SET(&rx_buf.lkey_ctrl, EFA_IO_RX_DESC_LAST, 0);
 
 		for (i = 0; i < wr->num_sge; i++) {
 			/* Set last indication if need) */
 			if (i == wr->num_sge - 1)
-				set_efa_io_rx_desc_last(&rx_buf, 1);
+				EFA_SET(&rx_buf.lkey_ctrl, EFA_IO_RX_DESC_LAST,
+					1);
 
 			addr = wr->sg_list[i].addr;
 
 			/* Set RX buffer desc from SGE */
 			rx_buf.length = wr->sg_list[i].length;
-			set_efa_io_rx_desc_lkey(&rx_buf, wr->sg_list[i].lkey);
+			EFA_SET(&rx_buf.lkey_ctrl, EFA_IO_RX_DESC_LKEY,
+				wr->sg_list[i].lkey);
 			rx_buf.buf_addr_lo = addr;
 			rx_buf.buf_addr_hi = (uint64_t)addr >> 32;
 
@@ -1078,6 +1621,25 @@ ring_db:
 	return err;
 }
 
+int efadv_query_ah(struct ibv_ah *ibvah, struct efadv_ah_attr *attr,
+		   uint32_t inlen)
+{
+	uint64_t comp_mask_out = 0;
+
+	if (!is_efa_dev(ibvah->context->device))
+		return EOPNOTSUPP;
+
+	if (!vext_field_avail(typeof(*attr), ahn, inlen))
+		return EINVAL;
+
+	memset(attr, 0, inlen);
+	attr->ahn = to_efa_ah(ibvah)->efa_ah;
+
+	attr->comp_mask = comp_mask_out;
+
+	return 0;
+}
+
 struct ibv_ah *efa_create_ah(struct ibv_pd *ibvpd, struct ibv_ah_attr *attr)
 {
 	struct efa_create_ah_resp resp = {};
@@ -1092,6 +1654,7 @@ struct ibv_ah *efa_create_ah(struct ibv_pd *ibvpd, struct ibv_ah_attr *attr)
 				&resp.ibv_resp, sizeof(resp));
 	if (err) {
 		free(ah);
+		errno = err;
 		return NULL;
 	}
 
