@@ -68,6 +68,9 @@
 #define HNS_ROCE_TPTR_OFFSET		0x1000
 #define HNS_ROCE_STATIC_RATE		3 /* Gbps */
 
+#define HNS_ROCE_ADDRESS_MASK 0xFFFFFFFF
+#define HNS_ROCE_ADDRESS_SHIFT 32
+
 #define roce_get_field(origin, mask, shift) \
 	(((le32toh(origin)) & (mask)) >> (shift))
 
@@ -83,6 +86,8 @@
 #define roce_set_bit(origin, shift, val) \
 	roce_set_field((origin), (1ul << (shift)), (shift), (val))
 
+#define hr_ilog32(n)		ilog32((n) - 1)
+
 enum {
 	HNS_ROCE_QP_TABLE_BITS		= 8,
 	HNS_ROCE_QP_TABLE_SIZE		= 1 << HNS_ROCE_QP_TABLE_BITS,
@@ -108,6 +113,7 @@ struct hns_roce_buf {
 };
 
 #define BIT_CNT_PER_BYTE       8
+#define BIT_CNT_PER_U64		64
 
 /* the sw doorbell type; */
 enum hns_roce_db_type {
@@ -181,9 +187,9 @@ struct hns_roce_srq {
 	pthread_spinlock_t		lock;
 	unsigned long			*wrid;
 	unsigned int			srqn;
-	int				max;
+	unsigned int			max_wqe;
 	unsigned int			max_gs;
-	int				wqe_shift;
+	unsigned int			wqe_shift;
 	int				head;
 	int				tail;
 	unsigned int			*db;
@@ -199,10 +205,18 @@ struct hns_roce_wq {
 	unsigned int			head;
 	unsigned int			tail;
 	unsigned int			max_gs;
-	int				wqe_shift;
+	unsigned int			wqe_shift;
+	unsigned int			shift; /* wq size is 2^shift */
 	int				offset;
 };
 
+/* record the result of sge process */
+struct hns_roce_sge_info {
+	unsigned int valid_num; /* sge length is not 0 */
+	unsigned int start_idx; /* start position of extend sge */
+	unsigned int total_len; /* total length of valid sges */
+};
+
 struct hns_roce_sge_ex {
 	int				offset;
 	unsigned int			sge_cnt;
@@ -234,7 +248,7 @@ struct hns_roce_qp {
 	struct hns_roce_wq		rq;
 	unsigned int			*rdb;
 	unsigned int			*sdb;
-	struct hns_roce_sge_ex		sge;
+	struct hns_roce_sge_ex		ex_sge;
 	unsigned int			next_sge;
 	int				port_num;
 	int				sl;
@@ -248,11 +262,6 @@ struct hns_roce_u_hw {
 	struct verbs_context_ops hw_ops;
 };
 
-static inline unsigned long align(unsigned long val, unsigned long align)
-{
-	return (val + align - 1) & ~(align - 1);
-}
-
 static inline struct hns_roce_device *to_hr_dev(struct ibv_device *ibv_dev)
 {
 	return container_of(ibv_dev, struct hns_roce_device, ibv_dev.device);
