@@ -52,6 +52,8 @@ enum {
 /* V2 REG DEFINITION */
 #define ROCEE_VF_DB_CFG0_OFFSET			0x0230
 
+#define HNS_ROCE_IDX_QUE_ENTRY_SZ		4
+
 enum {
 	HNS_ROCE_WQE_OP_SEND = 0x0,
 	HNS_ROCE_WQE_OP_SEND_WITH_INV = 0x1,
@@ -100,7 +102,6 @@ enum {
 };
 
 enum {
-	HNS_ROCE_V2_CQE_QPN_MASK	= 0x3ffff,
 	HNS_ROCE_V2_CQE_STATUS_MASK	= 0xff,
 	HNS_ROCE_V2_CQE_OPCODE_MASK	= 0x1f,
 };
@@ -122,6 +123,14 @@ enum {
 	HNS_ROCE_V2_CQE_REMOTE_ABORTED_ERR		= 0x22,
 };
 
+enum {
+	HNS_ROCE_V2_SQ_DB,
+	HNS_ROCE_V2_RQ_DB,
+	HNS_ROCE_V2_SRQ_DB,
+	HNS_ROCE_V2_CQ_DB_PTR,
+	HNS_ROCE_V2_CQ_DB_NTR,
+};
+
 struct hns_roce_db {
 	__le32	byte_4;
 	__le32	parameter;
@@ -148,26 +157,15 @@ struct hns_roce_db {
 #define DB_PARAM_SL_M \
 	(((1UL << 3) - 1) << DB_PARAM_SL_S)
 
-struct hns_roce_v2_cq_db {
-	__le32	byte_4;
-	__le32	parameter;
-};
+#define DB_PARAM_CQ_CONSUMER_IDX_S 0
+#define DB_PARAM_CQ_CONSUMER_IDX_M \
+	(((1UL << 24) - 1) << DB_PARAM_CQ_CONSUMER_IDX_S)
 
-#define CQ_DB_BYTE_4_TAG_S 0
-#define CQ_DB_BYTE_4_TAG_M   (((1UL << 23) - 1) << CQ_DB_BYTE_4_TAG_S)
+#define DB_PARAM_CQ_NOTIFY_S 24
 
-#define CQ_DB_BYTE_4_CMD_S 24
-#define CQ_DB_BYTE_4_CMD_M   (((1UL << 4) - 1) << CQ_DB_BYTE_4_CMD_S)
-
-#define CQ_DB_PARAMETER_CQ_CONSUMER_IDX_S 0
-#define CQ_DB_PARAMETER_CQ_CONSUMER_IDX_M \
-	(((1UL << 24) - 1) << CQ_DB_PARAMETER_CQ_CONSUMER_IDX_S)
-
-#define CQ_DB_PARAMETER_NOTIFY_S 24
-
-#define CQ_DB_PARAMETER_CMD_SN_S 25
-#define CQ_DB_PARAMETER_CMD_SN_M \
-	(((1UL << 2) - 1) << CQ_DB_PARAMETER_CMD_SN_S)
+#define DB_PARAM_CQ_CMD_SN_S 25
+#define DB_PARAM_CQ_CMD_SN_M \
+	(((1UL << 2) - 1) << DB_PARAM_CQ_CMD_SN_S)
 
 struct hns_roce_v2_cqe {
 	__le32	byte_4;
@@ -218,7 +216,7 @@ struct hns_roce_v2_cqe {
 #define CQE_BYTE_32_PORTN_S 27
 #define CQE_BYTE_32_PORTN_M   (((1UL << 3) - 1) << CQE_BYTE_32_PORTN_S)
 
-#define CQE_BYTE_32_GLH_S 30
+#define CQE_BYTE_32_GRH_S 30
 
 #define CQE_BYTE_32_LPK_S 31
 
@@ -292,4 +290,10 @@ struct hns_roce_wqe_atomic_seg {
 int hns_roce_u_v2_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 			    struct ibv_send_wr **bad_wr);
 
+#define DATA_TYPE_NUM 2
+#define STANDARD_ATOMIC_U_BYTE_8 0x8
+#define EXTEND_ATOMIC_U_BYTE_16 0x10
+#define EXTEND_ATOMIC_U_BYTE_32 0x20
+#define EXTEND_ATOMIC_U_BYTE_64 0x40
+
 #endif /* _HNS_ROCE_U_HW_V2_H */
