@@ -54,12 +54,12 @@ struct dr_qp_init_attr {
 	struct ibv_qp_cap	cap;
 };
 
-static inline void *dr_cq_get_cqe(struct dr_cq *dr_cq, int n)
+static void *dr_cq_get_cqe(struct dr_cq *dr_cq, int n)
 {
 	return dr_cq->buf + n * dr_cq->cqe_sz;
 }
 
-static inline void *dr_cq_get_sw_cqe(struct dr_cq *dr_cq, int n)
+static void *dr_cq_get_sw_cqe(struct dr_cq *dr_cq, int n)
 {
 	void *cqe = dr_cq_get_cqe(dr_cq, n & (dr_cq->ncqe - 1));
 	struct mlx5_cqe64 *cqe64;
@@ -74,8 +74,8 @@ static inline void *dr_cq_get_sw_cqe(struct dr_cq *dr_cq, int n)
 		return NULL;
 }
 
-static inline int dr_get_next_cqe(struct dr_cq *dr_cq,
-				  struct mlx5_cqe64 **pcqe64)
+static int dr_get_next_cqe(struct dr_cq *dr_cq,
+			   struct mlx5_cqe64 **pcqe64)
 {
 	struct mlx5_cqe64 *cqe64;
 
@@ -95,7 +95,7 @@ static inline int dr_get_next_cqe(struct dr_cq *dr_cq,
 	return CQ_OK;
 }
 
-static inline int dr_parse_cqe(struct dr_cq *dr_cq, struct mlx5_cqe64 *cqe64)
+static int dr_parse_cqe(struct dr_cq *dr_cq, struct mlx5_cqe64 *cqe64)
 {
 	uint16_t wqe_ctr;
 	uint8_t opcode;
@@ -118,7 +118,7 @@ static inline int dr_parse_cqe(struct dr_cq *dr_cq, struct mlx5_cqe64 *cqe64)
 	return CQ_POLL_ERR;
 }
 
-static inline int dr_cq_poll_one(struct dr_cq *dr_cq)
+static int dr_cq_poll_one(struct dr_cq *dr_cq)
 {
 	struct mlx5_cqe64 *cqe64;
 	int err;
@@ -179,7 +179,7 @@ static int dr_calc_sq_size(struct dr_qp *dr_qp,
 
 	wq_size = roundup_pow_of_two(attr->cap.max_send_wr * wqe_size);
 	dr_qp->sq.wqe_cnt = wq_size / MLX5_SEND_WQE_BB;
-	dr_qp->sq.wqe_shift = mlx5_ilog2(MLX5_SEND_WQE_BB);
+	dr_qp->sq.wqe_shift = STATIC_ILOG_32(MLX5_SEND_WQE_BB) - 1;
 	dr_qp->sq.max_gs = attr->cap.max_send_sge;
 	dr_qp->sq.max_post = wq_size / wqe_size;
 
@@ -210,8 +210,8 @@ static int dr_calc_rq_size(struct dr_qp *dr_qp,
 	wq_size = roundup_pow_of_two(attr->cap.max_recv_wr) * wqe_size;
 	wq_size = max(wq_size, MLX5_SEND_WQE_BB);
 	dr_qp->rq.wqe_cnt = wq_size / wqe_size;
-	dr_qp->rq.wqe_shift = mlx5_ilog2(wqe_size);
-	dr_qp->rq.max_post = 1 << mlx5_ilog2(wq_size / wqe_size);
+	dr_qp->rq.wqe_shift = ilog32(wqe_size - 1);
+	dr_qp->rq.max_post = 1 << ilog32(wq_size / wqe_size - 1);
 	dr_qp->rq.max_gs = wqe_size / sizeof(struct mlx5_wqe_data_seg);
 
 	return wq_size;
@@ -371,15 +371,15 @@ static int dr_destroy_qp(struct dr_qp *dr_qp)
 	return 0;
 }
 
-static inline void dr_set_raddr_seg(struct mlx5_wqe_raddr_seg *rseg,
-				    uint64_t remote_addr, uint32_t rkey)
+static void dr_set_raddr_seg(struct mlx5_wqe_raddr_seg *rseg,
+			     uint64_t remote_addr, uint32_t rkey)
 {
 	rseg->raddr    = htobe64(remote_addr);
 	rseg->rkey     = htobe32(rkey);
 	rseg->reserved = 0;
 }
 
-static inline void dr_post_send_db(struct dr_qp *dr_qp, int size, void *ctrl)
+static void dr_post_send_db(struct dr_qp *dr_qp, int size, void *ctrl)
 {
 	dr_qp->sq.head += 2; /* RDMA_WRITE + RDMA_READ */
 
@@ -406,9 +406,9 @@ static void dr_set_data_ptr_seg(struct mlx5_wqe_data_seg *dseg,
 	dseg->addr       = htobe64(data_seg->addr);
 }
 
-static inline int dr_set_data_inl_seg(struct dr_qp *dr_qp,
-				      struct dr_data_seg *data_seg,
-				      void *wqe, uint32_t opcode, int *sz)
+static int dr_set_data_inl_seg(struct dr_qp *dr_qp,
+			       struct dr_data_seg *data_seg,
+			       void *wqe, uint32_t opcode, int *sz)
 {
 	struct mlx5_wqe_inline_seg *seg;
 	void *qend = dr_qp->sq.qend;
@@ -716,7 +716,7 @@ int dr_send_postsend_htbl(struct mlx5dv_dr_domain *dmn, struct dr_ste_htbl *htbl
 
 	/* Send the data iteration times */
 	for (i = 0; i < iterations; i++) {
-		uint32_t ste_index = (i > 0) ? (i * (byte_size / DR_STE_SIZE)) - 1 : 0;
+		uint32_t ste_index = i * (byte_size / DR_STE_SIZE);
 		struct postsend_info send_info = {};
 
 		/* Copy all ste's on the data buffer, need to add the bit_mask */
@@ -781,7 +781,7 @@ int dr_send_postsend_formated_htbl(struct mlx5dv_dr_domain *dmn,
 
 	/* Send the data iteration times */
 	for (i = 0; i < iterations; i++) {
-		uint32_t ste_index = (i > 0) ? (i * (byte_size / DR_STE_SIZE)) - 1 : 0;
+		uint32_t ste_index = i * (byte_size / DR_STE_SIZE);
 		struct postsend_info send_info = {};
 
 		send_info.write.addr	= (uintptr_t) data;
