@@ -566,7 +566,7 @@ bool dr_ste_not_used_ste(struct dr_ste *ste)
 	return !atomic_load(&ste->refcount);
 }
 
-static inline uint16_t get_bits_per_mask(uint16_t byte_mask)
+static uint16_t get_bits_per_mask(uint16_t byte_mask)
 {
 	uint16_t bits = 0;
 
@@ -1036,6 +1036,9 @@ static void dr_ste_copy_mask_misc3(char *mask, struct dr_match_misc3 *spec)
 	spec->icmpv4_code = DEVX_GET(dr_match_set_misc3, mask, icmp_code);
 	spec->icmpv6_type = DEVX_GET(dr_match_set_misc3, mask, icmpv6_type);
 	spec->icmpv6_code = DEVX_GET(dr_match_set_misc3, mask, icmpv6_code);
+	spec->gtpu_flags    = DEVX_GET(dr_match_set_misc3, mask, gtpu_flags);
+	spec->gtpu_msg_type = DEVX_GET(dr_match_set_misc3, mask, gtpu_msg_type);
+	spec->gtpu_teid     = DEVX_GET(dr_match_set_misc3, mask, gtpu_teid);
 }
 
 #define MAX_PARAM_SIZE 512
@@ -1050,7 +1053,7 @@ void dr_ste_copy_param(uint8_t match_criteria,
 	void *buff;
 
 	if (match_criteria & DR_MATCHER_CRITERIA_OUTER) {
-		if (mask->match_sz < sizeof(struct dr_match_spec)) {
+		if (mask->match_sz < DEVX_ST_SZ_BYTES(dr_match_spec)) {
 			memcpy(tail_param, data, mask->match_sz);
 			buff = tail_param;
 		} else {
@@ -1058,11 +1061,11 @@ void dr_ste_copy_param(uint8_t match_criteria,
 		}
 		dr_ste_copy_mask_spec(buff, &set_param->outer);
 	}
-	param_location = sizeof(struct dr_match_spec);
+	param_location = DEVX_ST_SZ_BYTES(dr_match_spec);
 
 	if (match_criteria & DR_MATCHER_CRITERIA_MISC) {
 		if (mask->match_sz < param_location +
-		    sizeof(struct dr_match_misc)) {
+		    DEVX_ST_SZ_BYTES(dr_match_set_misc)) {
 			memcpy(tail_param, data + param_location,
 			       mask->match_sz - param_location);
 			buff = tail_param;
@@ -1071,11 +1074,11 @@ void dr_ste_copy_param(uint8_t match_criteria,
 		}
 		dr_ste_copy_mask_misc(buff, &set_param->misc);
 	}
-	param_location += sizeof(struct dr_match_misc);
+	param_location += DEVX_ST_SZ_BYTES(dr_match_set_misc);
 
 	if (match_criteria & DR_MATCHER_CRITERIA_INNER) {
 		if (mask->match_sz < param_location +
-		    sizeof(struct dr_match_spec)) {
+		    DEVX_ST_SZ_BYTES(dr_match_spec)) {
 			memcpy(tail_param, data + param_location,
 			       mask->match_sz - param_location);
 			buff = tail_param;
@@ -1084,11 +1087,11 @@ void dr_ste_copy_param(uint8_t match_criteria,
 		}
 		dr_ste_copy_mask_spec(buff, &set_param->inner);
 	}
-	param_location += sizeof(struct dr_match_spec);
+	param_location += DEVX_ST_SZ_BYTES(dr_match_spec);
 
 	if (match_criteria & DR_MATCHER_CRITERIA_MISC2) {
 		if (mask->match_sz < param_location +
-		    sizeof(struct dr_match_misc2)) {
+		    DEVX_ST_SZ_BYTES(dr_match_set_misc2)) {
 			memcpy(tail_param, data + param_location,
 			       mask->match_sz - param_location);
 			buff = tail_param;
@@ -1098,11 +1101,11 @@ void dr_ste_copy_param(uint8_t match_criteria,
 		dr_ste_copy_mask_misc2(buff, &set_param->misc2);
 	}
 
-	param_location += sizeof(struct dr_match_misc2);
+	param_location += DEVX_ST_SZ_BYTES(dr_match_set_misc2);
 
 	if (match_criteria & DR_MATCHER_CRITERIA_MISC3) {
 		if (mask->match_sz < param_location +
-		    sizeof(struct dr_match_misc3)) {
+		    DEVX_ST_SZ_BYTES(dr_match_set_misc3)) {
 			memcpy(tail_param, data + param_location,
 			       mask->match_sz - param_location);
 			buff = tail_param;
@@ -2125,68 +2128,160 @@ void dr_ste_build_eth_l4_misc(struct dr_ste_build *sb,
 	sb->ste_build_tag_func = &dr_ste_build_eth_l4_misc_tag;
 }
 
-static void dr_ste_build_flex_parser_tnl_bit_mask(struct dr_match_param *value,
-						  bool inner, uint8_t *bit_mask)
+static void
+dr_ste_build_flex_parser_tnl_vxlan_gpe_bit_mask(struct dr_match_param *value,
+						bool inner, uint8_t *bit_mask)
 {
 	struct dr_match_misc3 *misc_3_mask = &value->misc3;
 
-	if (misc_3_mask->outer_vxlan_gpe_flags ||
-	    misc_3_mask->outer_vxlan_gpe_next_protocol) {
-		DR_STE_SET(flex_parser_tnl, bit_mask,
-			   flex_parser_tunneling_header_63_32,
-			   (misc_3_mask->outer_vxlan_gpe_flags << 24) |
-			   (misc_3_mask->outer_vxlan_gpe_next_protocol));
-		misc_3_mask->outer_vxlan_gpe_flags = 0;
-		misc_3_mask->outer_vxlan_gpe_next_protocol = 0;
-	}
-
-	if (misc_3_mask->outer_vxlan_gpe_vni) {
-		DR_STE_SET(flex_parser_tnl, bit_mask,
-			   flex_parser_tunneling_header_31_0,
-			   misc_3_mask->outer_vxlan_gpe_vni << 8);
-		misc_3_mask->outer_vxlan_gpe_vni = 0;
-	}
+	DR_STE_SET_MASK_V(flex_parser_tnl_vxlan_gpe, bit_mask,
+			  outer_vxlan_gpe_flags,
+			  misc_3_mask, outer_vxlan_gpe_flags);
+	DR_STE_SET_MASK_V(flex_parser_tnl_vxlan_gpe, bit_mask,
+			  outer_vxlan_gpe_next_protocol,
+			  misc_3_mask, outer_vxlan_gpe_next_protocol);
+	DR_STE_SET_MASK_V(flex_parser_tnl_vxlan_gpe, bit_mask,
+			  outer_vxlan_gpe_vni,
+			  misc_3_mask, outer_vxlan_gpe_vni);
 }
 
-static int dr_ste_build_flex_parser_tnl_tag(struct dr_match_param *value,
-					    struct dr_ste_build *sb,
-					    uint8_t *hw_ste_p)
+static int
+dr_ste_build_flex_parser_tnl_vxlan_gpe_tag(struct dr_match_param *value,
+					   struct dr_ste_build *sb,
+					   uint8_t *hw_ste_p)
 {
 	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
 	struct dr_match_misc3 *misc3 = &value->misc3;
 	uint8_t *tag = hw_ste->tag;
 
-	if (misc3->outer_vxlan_gpe_flags ||
-	    misc3->outer_vxlan_gpe_next_protocol) {
-		DR_STE_SET(flex_parser_tnl, tag,
-			   flex_parser_tunneling_header_63_32,
-			   (misc3->outer_vxlan_gpe_flags << 24) |
-			   (misc3->outer_vxlan_gpe_next_protocol));
-		misc3->outer_vxlan_gpe_flags = 0;
-		misc3->outer_vxlan_gpe_next_protocol = 0;
-	}
-
-	if (misc3->outer_vxlan_gpe_vni) {
-		DR_STE_SET(flex_parser_tnl, tag,
-			   flex_parser_tunneling_header_31_0,
-			   misc3->outer_vxlan_gpe_vni << 8);
-		misc3->outer_vxlan_gpe_vni = 0;
-	}
+	DR_STE_SET_TAG(flex_parser_tnl_vxlan_gpe, tag,
+		       outer_vxlan_gpe_flags, misc3,
+		       outer_vxlan_gpe_flags);
+	DR_STE_SET_TAG(flex_parser_tnl_vxlan_gpe, tag,
+		       outer_vxlan_gpe_next_protocol, misc3,
+		       outer_vxlan_gpe_next_protocol);
+	DR_STE_SET_TAG(flex_parser_tnl_vxlan_gpe, tag,
+		       outer_vxlan_gpe_vni, misc3,
+		       outer_vxlan_gpe_vni);
 
 	return 0;
 }
 
-void dr_ste_build_flex_parser_tnl(struct dr_ste_build *sb,
-				  struct dr_match_param *mask,
-				  bool inner, bool rx)
+void dr_ste_build_flex_parser_tnl_vxlan_gpe(struct dr_ste_build *sb,
+					    struct dr_match_param *mask,
+					    bool inner, bool rx)
 {
-	dr_ste_build_flex_parser_tnl_bit_mask(mask, inner, sb->bit_mask);
-
+	dr_ste_build_flex_parser_tnl_vxlan_gpe_bit_mask(mask, inner,
+							sb->bit_mask);
 	sb->rx = rx;
 	sb->inner = inner;
 	sb->lu_type = DR_STE_LU_TYPE_FLEX_PARSER_TNL_HEADER;
 	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
-	sb->ste_build_tag_func = &dr_ste_build_flex_parser_tnl_tag;
+	sb->ste_build_tag_func = &dr_ste_build_flex_parser_tnl_vxlan_gpe_tag;
+}
+
+static void
+dr_ste_build_flex_parser_tnl_geneve_bit_mask(struct dr_match_param *value,
+					     uint8_t *bit_mask)
+{
+	struct dr_match_misc *misc_mask = &value->misc;
+
+	DR_STE_SET_MASK_V(flex_parser_tnl_geneve, bit_mask,
+			  geneve_protocol_type,
+			  misc_mask, geneve_protocol_type);
+	DR_STE_SET_MASK_V(flex_parser_tnl_geneve, bit_mask,
+			  geneve_oam,
+			  misc_mask, geneve_oam);
+	DR_STE_SET_MASK_V(flex_parser_tnl_geneve, bit_mask,
+			  geneve_opt_len,
+			  misc_mask, geneve_opt_len);
+	DR_STE_SET_MASK_V(flex_parser_tnl_geneve, bit_mask,
+			  geneve_vni,
+			  misc_mask, geneve_vni);
+}
+
+static int
+dr_ste_build_flex_parser_tnl_geneve_tag(struct dr_match_param *value,
+					struct dr_ste_build *sb,
+					uint8_t *hw_ste_p)
+{
+	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
+	struct dr_match_misc *misc = &value->misc;
+	uint8_t *tag = hw_ste->tag;
+
+	DR_STE_SET_TAG(flex_parser_tnl_geneve, tag,
+		       geneve_protocol_type, misc, geneve_protocol_type);
+	DR_STE_SET_TAG(flex_parser_tnl_geneve, tag,
+		       geneve_oam, misc, geneve_oam);
+	DR_STE_SET_TAG(flex_parser_tnl_geneve, tag,
+		       geneve_opt_len, misc, geneve_opt_len);
+	DR_STE_SET_TAG(flex_parser_tnl_geneve, tag,
+		       geneve_vni, misc, geneve_vni);
+
+	return 0;
+}
+
+void dr_ste_build_flex_parser_tnl_geneve(struct dr_ste_build *sb,
+					 struct dr_match_param *mask,
+					 bool inner, bool rx)
+{
+	dr_ste_build_flex_parser_tnl_geneve_bit_mask(mask, sb->bit_mask);
+	sb->rx = rx;
+	sb->inner = inner;
+	sb->lu_type = DR_STE_LU_TYPE_FLEX_PARSER_TNL_HEADER;
+	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
+	sb->ste_build_tag_func = &dr_ste_build_flex_parser_tnl_geneve_tag;
+}
+
+static void
+dr_ste_build_flex_parser_tnl_gtpu_bit_mask(struct dr_match_param *value,
+					   uint8_t *bit_mask)
+{
+	struct dr_match_misc3 *misc3 = &value->misc3;
+
+	DR_STE_SET_MASK_V(flex_parser_tnl_gtpu, bit_mask,
+			  gtpu_flags, misc3,
+			  gtpu_flags);
+	DR_STE_SET_MASK_V(flex_parser_tnl_gtpu, bit_mask,
+			  gtpu_msg_type, misc3,
+			  gtpu_msg_type);
+	DR_STE_SET_MASK_V(flex_parser_tnl_gtpu, bit_mask,
+			  gtpu_teid, misc3,
+			  gtpu_teid);
+}
+
+static int
+dr_ste_build_flex_parser_tnl_gtpu_tag(struct dr_match_param *value,
+				      struct dr_ste_build *sb,
+				      uint8_t *hw_ste_p)
+{
+	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
+	struct dr_match_misc3 *misc3 = &value->misc3;
+	uint8_t *tag = hw_ste->tag;
+
+	DR_STE_SET_TAG(flex_parser_tnl_gtpu, tag,
+		       gtpu_flags, misc3,
+		       gtpu_flags);
+	DR_STE_SET_TAG(flex_parser_tnl_gtpu, tag,
+		       gtpu_msg_type, misc3,
+		       gtpu_msg_type);
+	DR_STE_SET_TAG(flex_parser_tnl_gtpu, tag,
+		       gtpu_teid, misc3,
+		       gtpu_teid);
+
+	return 0;
+}
+
+void dr_ste_build_flex_parser_tnl_gtpu(struct dr_ste_build *sb,
+				       struct dr_match_param *mask,
+				       bool inner, bool rx)
+{
+	dr_ste_build_flex_parser_tnl_gtpu_bit_mask(mask, sb->bit_mask);
+	sb->rx = rx;
+	sb->inner = inner;
+	sb->lu_type = DR_STE_LU_TYPE_FLEX_PARSER_TNL_HEADER;
+	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
+	sb->ste_build_tag_func = &dr_ste_build_flex_parser_tnl_gtpu_tag;
 }
 
 static void dr_ste_build_register_0_bit_mask(struct dr_match_param *value,
@@ -2282,7 +2377,7 @@ static int dr_ste_build_src_gvmi_qpn_bit_mask(struct dr_match_param *value,
 {
 	struct dr_match_misc *misc_mask = &value->misc;
 
-	if (misc_mask->source_port != 0xffff) {
+	if (misc_mask->source_port && misc_mask->source_port != 0xffff) {
 		errno = EINVAL;
 		return errno;
 	}
