@@ -447,7 +447,24 @@ struct mlx5_ifc_dr_match_set_misc3_bits {
 	u8         icmpv6_type[0x8];
 	u8         icmpv6_code[0x8];
 
-	u8         reserved_at_120[0xe0];
+	u8         reserved_at_120[0x20];
+
+	u8         gtpu_teid[0x20];
+
+	u8         gtpu_msg_type[0x8];
+	u8         reserved_at_148[0x5];
+	u8         gtpu_flags[0x3];
+	u8         reserved_at_150[0x10];
+
+	u8         reserved_at_160[0x80];
+};
+
+struct mlx5_ifc_dr_match_param_bits {
+	struct mlx5_ifc_dr_match_spec_bits outer;
+	struct mlx5_ifc_dr_match_set_misc_bits misc;
+	struct mlx5_ifc_dr_match_spec_bits inner;
+	struct mlx5_ifc_dr_match_set_misc2_bits misc2;
+	struct mlx5_ifc_dr_match_set_misc3_bits misc3;
 };
 
 struct mlx5_ifc_flow_table_prop_layout_bits {
@@ -502,9 +519,11 @@ struct mlx5_ifc_flow_table_prop_layout_bits {
 };
 
 enum {
+	MLX5_FLEX_PARSER_GENEVE_ENABLED		= 1 << 3,
 	MLX5_FLEX_PARSER_VXLAN_GPE_ENABLED	= 1 << 7,
 	MLX5_FLEX_PARSER_ICMP_V4_ENABLED	= 1 << 8,
 	MLX5_FLEX_PARSER_ICMP_V6_ENABLED	= 1 << 9,
+	MLX5_FLEX_PARSER_GTPU_ENABLED		= 1 << 11,
 };
 
 struct mlx5_ifc_cmd_hca_cap_bits {
@@ -972,12 +991,44 @@ struct mlx5_ifc_flow_table_eswitch_cap_bits {
 	u8      reserved_at_1900[0x6700];
 };
 
+struct mlx5_ifc_odp_per_transport_service_cap_bits {
+	u8         send[0x1];
+	u8         receive[0x1];
+	u8         write[0x1];
+	u8         read[0x1];
+	u8         atomic[0x1];
+	u8         srq_receive[0x1];
+	u8         reserved_at_6[0x1a];
+};
+
+struct mlx5_ifc_odp_cap_bits {
+	u8         reserved_at_0[0x40];
+
+	u8         sig[0x1];
+	u8         reserved_at_41[0x1f];
+
+	u8         reserved_at_60[0x20];
+
+	struct mlx5_ifc_odp_per_transport_service_cap_bits rc_odp_caps;
+
+	struct mlx5_ifc_odp_per_transport_service_cap_bits uc_odp_caps;
+
+	struct mlx5_ifc_odp_per_transport_service_cap_bits ud_odp_caps;
+
+	struct mlx5_ifc_odp_per_transport_service_cap_bits xrc_odp_caps;
+
+	struct mlx5_ifc_odp_per_transport_service_cap_bits dc_odp_caps;
+
+	u8         reserved_at_120[0x6e0];
+};
+
 union mlx5_ifc_hca_cap_union_bits {
 	struct mlx5_ifc_atomic_caps_bits atomic_caps;
 	struct mlx5_ifc_cmd_hca_cap_bits cmd_hca_cap;
 	struct mlx5_ifc_flow_table_nic_cap_bits flow_table_nic_cap;
 	struct mlx5_ifc_flow_table_eswitch_cap_bits flow_table_eswitch_cap;
 	struct mlx5_ifc_device_mem_cap_bits device_mem_cap;
+	struct mlx5_ifc_odp_cap_bits odp_cap;
 	u8         reserved_at_0[0x8000];
 };
 
@@ -1007,6 +1058,7 @@ struct mlx5_ifc_query_hca_cap_in_bits {
 };
 
 enum mlx5_cap_type {
+	MLX5_CAP_ODP = 2,
 	MLX5_CAP_ATOMIC = 3,
 };
 
@@ -1147,36 +1199,43 @@ struct mlx5_ifc_ste_sx_transmit_bits {
 	u8         reserved_at_4[0x4];
 	u8         entry_sub_type[0x8];
 	u8         byte_mask[0x10];
+
 	u8         next_table_base_63_48[0x10];
 	u8         next_lu_type[0x8];
 	u8         next_table_base_39_32_size[0x8];
+
 	u8         next_table_base_31_5_size[0x1b];
 	u8         linear_hash_enable[0x1];
 	u8         reserved_at_5c[0x2];
 	u8         next_table_rank[0x2];
-	u8         sx_wire;
-	u8         sx_func_lb;
-	u8         sx_sniffer;
-	u8         sx_wire_enable;
-	u8         sx_func_lb_enable;
-	u8         sx_sniffer_enable;
-	u8         action_type[3];
-	u8         reserved_at_69;
-	u8         action_description[6];
-	u8         gvmi[16];
-	u8         encap_pointer_vlan_data[32];
-	u8         loopback_syndome_en[8];
-	u8         loopback_syndome[8];
-	u8         counter_trigger[16];
-	u8         miss_address_63_48[16];
-	u8         counter_trigger_23_16[8];
-	u8         miss_address_39_32[8];
-	u8         miss_address_31_6[26];
-	u8         learning_point;
-	u8         go_back;
-	u8         match_polarity;
-	u8         mask_mode;
-	u8         miss_rank[2];
+
+	u8         sx_wire[0x1];
+	u8         sx_func_lb[0x1];
+	u8         sx_sniffer[0x1];
+	u8         sx_wire_enable[0x1];
+	u8         sx_func_lb_enable[0x1];
+	u8         sx_sniffer_enable[0x1];
+	u8         action_type[0x3];
+	u8         reserved_at_69[0x1];
+	u8         action_description[0x6];
+	u8         gvmi[0x10];
+
+	u8         encap_pointer_vlan_data[0x20];
+
+	u8         loopback_syndome_en[0x8];
+	u8         loopback_syndome[0x8];
+	u8         counter_trigger[0x10];
+
+	u8         miss_address_63_48[0x10];
+	u8         counter_trigger_23_16[0x8];
+	u8         miss_address_39_32[0x8];
+
+	u8         miss_address_31_6[0x1a];
+	u8         learning_point[0x1];
+	u8         go_back[0x1];
+	u8         match_polarity[0x1];
+	u8         mask_mode[0x1];
+	u8         miss_rank[0x2];
 };
 
 struct mlx5_ifc_ste_rx_steering_mult_bits {
@@ -1184,24 +1243,31 @@ struct mlx5_ifc_ste_rx_steering_mult_bits {
 	u8         reserved_at_4[0x4];
 	u8         entry_sub_type[0x8];
 	u8         byte_mask[0x10];
+
 	u8         next_table_base_63_48[0x10];
 	u8         next_lu_type[0x8];
 	u8         next_table_base_39_32_size[0x8];
+
 	u8         next_table_base_31_5_size[0x1b];
 	u8         linear_hash_enable[0x1];
-	u8         reserved_at_[0x2];
+	u8         reserved_at_5c[0x2];
 	u8         next_table_rank[0x2];
+
 	u8         member_count[0x10];
 	u8         gvmi[0x10];
+
 	u8         qp_list_pointer[0x20];
+
 	u8         reserved_at_a0[0x1];
 	u8         tunneling_action[0x3];
 	u8         action_description[0x4];
 	u8         reserved_at_a8[0x8];
 	u8         counter_trigger_15_0[0x10];
+
 	u8         miss_address_63_48[0x10];
 	u8         counter_trigger_23_16[0x08];
 	u8         miss_address_39_32[0x8];
+
 	u8         miss_address_31_6[0x1a];
 	u8         learning_point[0x1];
 	u8         fail_on_error[0x1];
@@ -1215,24 +1281,31 @@ struct mlx5_ifc_ste_modify_packet_bits {
 	u8         reserved_at_4[0x4];
 	u8         entry_sub_type[0x8];
 	u8         byte_mask[0x10];
+
 	u8         next_table_base_63_48[0x10];
 	u8         next_lu_type[0x8];
 	u8         next_table_base_39_32_size[0x8];
+
 	u8         next_table_base_31_5_size[0x1b];
 	u8         linear_hash_enable[0x1];
-	u8         reserved_at_[0x2];
+	u8         reserved_at_5c[0x2];
 	u8         next_table_rank[0x2];
+
 	u8         number_of_re_write_actions[0x10];
 	u8         gvmi[0x10];
+
 	u8         header_re_write_actions_pointer[0x20];
+
 	u8         reserved_at_a0[0x1];
 	u8         tunneling_action[0x3];
 	u8         action_description[0x4];
 	u8         reserved_at_a8[0x8];
 	u8         counter_trigger_15_0[0x10];
+
 	u8         miss_address_63_48[0x10];
 	u8         counter_trigger_23_16[0x08];
 	u8         miss_address_39_32[0x8];
+
 	u8         miss_address_31_6[0x1a];
 	u8         learning_point[0x1];
 	u8         fail_on_error[0x1];
@@ -1243,8 +1316,10 @@ struct mlx5_ifc_ste_modify_packet_bits {
 
 struct mlx5_ifc_ste_eth_l2_src_bits {
 	u8         smac_47_16[0x20];
+
 	u8         smac_15_0[0x10];
 	u8         l3_ethertype[0x10];
+
 	u8         qp_type[0x2];
 	u8         ethertype_filter[0x1];
 	u8         reserved_at_43[0x1];
@@ -1258,6 +1333,7 @@ struct mlx5_ifc_ste_eth_l2_src_bits {
 	u8         first_vlan_qualifier[0x2];
 	u8         reserved_at_52[0x2];
 	u8         first_vlan_id[0xc];
+
 	u8         ip_fragmented[0x1];
 	u8         tcp_syn[0x1];
 	u8         encp_type[0x2];
@@ -1273,8 +1349,10 @@ struct mlx5_ifc_ste_eth_l2_src_bits {
 
 struct mlx5_ifc_ste_eth_l2_dst_bits {
 	u8         dmac_47_16[0x20];
+
 	u8         dmac_15_0[0x10];
 	u8         l3_ethertype[0x10];
+
 	u8         qp_type[0x2];
 	u8         ethertype_filter[0x1];
 	u8         reserved_at_43[0x1];
@@ -1288,6 +1366,7 @@ struct mlx5_ifc_ste_eth_l2_dst_bits {
 	u8         first_vlan_qualifier[0x2];
 	u8         reserved_at_52[0x2];
 	u8         first_vlan_id[0xc];
+
 	u8         ip_fragmented[0x1];
 	u8         tcp_syn[0x1];
 	u8         encp_type[0x2];
@@ -1303,9 +1382,12 @@ struct mlx5_ifc_ste_eth_l2_dst_bits {
 
 struct mlx5_ifc_ste_eth_l2_src_dst_bits {
 	u8         dmac_47_16[0x20];
+
 	u8         dmac_15_0[0x10];
 	u8         smac_47_32[0x10];
+
 	u8         smac_31_0[0x20];
+
 	u8         sx_sniffer[0x1];
 	u8         force_lb[0x1];
 	u8         functional_lb[0x1];
@@ -1321,9 +1403,12 @@ struct mlx5_ifc_ste_eth_l2_src_dst_bits {
 
 struct mlx5_ifc_ste_eth_l3_ipv4_5_tuple_bits {
 	u8         destination_address[0x20];
+
 	u8         source_address[0x20];
+
 	u8         source_port[0x10];
 	u8         destination_port[0x10];
+
 	u8         fragmented[0x1];
 	u8         first_fragment[0x1];
 	u8         reserved_at_62[0x2];
@@ -1345,16 +1430,22 @@ struct mlx5_ifc_ste_eth_l3_ipv4_5_tuple_bits {
 
 struct mlx5_ifc_ste_eth_l3_ipv6_dst_bits {
 	u8         dst_ip_127_96[0x20];
+
 	u8         dst_ip_95_64[0x20];
+
 	u8         dst_ip_63_32[0x20];
+
 	u8         dst_ip_31_0[0x20];
 };
 
 struct mlx5_ifc_ste_eth_l2_tnl_bits {
 	u8         dmac_47_16[0x20];
+
 	u8         dmac_15_0[0x10];
 	u8         l3_ethertype[0x10];
+
 	u8         l2_tunneling_network_id[0x20];
+
 	u8         ip_fragmented[0x1];
 	u8         tcp_syn[0x1];
 	u8         encp_type[0x2];
@@ -1371,8 +1462,11 @@ struct mlx5_ifc_ste_eth_l2_tnl_bits {
 
 struct mlx5_ifc_ste_eth_l3_ipv6_src_bits {
 	u8         src_ip_127_96[0x20];
+
 	u8         src_ip_95_64[0x20];
+
 	u8         src_ip_63_32[0x20];
+
 	u8         src_ip_31_0[0x20];
 };
 
@@ -1381,12 +1475,15 @@ struct mlx5_ifc_ste_eth_l3_ipv4_misc_bits {
 	u8         ihl[0x4];
 	u8         reserved_at_8[0x8];
 	u8         total_length[0x10];
+
 	u8         identification[0x10];
 	u8         flags[0x3];
 	u8         fragment_offset[0xd];
+
 	u8         time_to_live[0x8];
 	u8         reserved_at_48[0x8];
 	u8         checksum[0x10];
+
 	u8         reserved_at_60[0x20];
 };
 
@@ -1396,6 +1493,7 @@ struct mlx5_ifc_ste_eth_l4_bits {
 	u8         reserved_at_2[0x6];
 	u8         protocol[0x8];
 	u8         dst_port[0x10];
+
 	u8         ipv6_version[0x4];
 	u8         reserved_at_24[0x1];
 	u8         ecn[0x2];
@@ -1409,10 +1507,12 @@ struct mlx5_ifc_ste_eth_l4_bits {
 	u8         tcp_syn[0x1];
 	u8         tcp_fin[0x1];
 	u8         src_port[0x10];
+
 	u8         ipv6_payload_length[0x10];
 	u8         ipv6_hop_limit[0x8];
 	u8         dscp[0x6];
 	u8         reserved_at_5e[0x2];
+
 	u8         tcp_data_offset[0x4];
 	u8         reserved_at_64[0x8];
 	u8         flow_label[0x14];
@@ -1421,8 +1521,11 @@ struct mlx5_ifc_ste_eth_l4_bits {
 struct mlx5_ifc_ste_eth_l4_misc_bits {
 	u8         checksum[0x10];
 	u8         length[0x10];
+
 	u8         seq_num[0x20];
+
 	u8         ack_num[0x20];
+
 	u8         urgent_pointer[0x10];
 	u8         window_size[0x10];
 };
@@ -1432,8 +1535,11 @@ struct mlx5_ifc_ste_mpls_bits {
 	u8         mpls0_exp[0x3];
 	u8         mpls0_s_bos[0x1];
 	u8         mpls0_ttl[0x8];
+
 	u8         mpls1_label[0x20];
+
 	u8         mpls2_label[0x20];
+
 	u8         reserved_at_60[0x16];
 	u8         mpls4_s_bit[0x1];
 	u8         mpls4_qualifier[0x1];
@@ -1449,21 +1555,27 @@ struct mlx5_ifc_ste_mpls_bits {
 
 struct mlx5_ifc_ste_register_0_bits {
 	u8         register_0_h[0x20];
+
 	u8         register_0_l[0x20];
+
 	u8         register_1_h[0x20];
+
 	u8         register_1_l[0x20];
 };
 
 struct mlx5_ifc_ste_register_1_bits {
 	u8         register_2_h[0x20];
+
 	u8         register_2_l[0x20];
+
 	u8         register_3_h[0x20];
+
 	u8         register_3_l[0x20];
 };
 
 struct mlx5_ifc_ste_gre_bits {
 	u8         gre_c_present[0x1];
-	u8         reserved_at_30[0x1];
+	u8         reserved_at_1[0x1];
 	u8         gre_k_present[0x1];
 	u8         gre_s_present[0x1];
 	u8         strict_src_route[0x1];
@@ -1471,10 +1583,13 @@ struct mlx5_ifc_ste_gre_bits {
 	u8         flags[0x5];
 	u8         version[0x3];
 	u8         gre_protocol[0x10];
+
 	u8         checksum[0x10];
 	u8         offset[0x10];
+
 	u8         gre_key_h[0x18];
 	u8         gre_key_l[0x8];
+
 	u8         seq_num[0x20];
 };
 
@@ -1483,28 +1598,66 @@ struct mlx5_ifc_ste_flex_parser_0_bits {
 	u8         parser_3_exp[0x3];
 	u8         parser_3_s_bos[0x1];
 	u8         parser_3_ttl[0x8];
+
 	u8         flex_parser_2[0x20];
+
 	u8         flex_parser_1[0x20];
+
 	u8         flex_parser_0[0x20];
 };
 
 struct mlx5_ifc_ste_flex_parser_1_bits {
 	u8         flex_parser_7[0x20];
+
 	u8         flex_parser_6[0x20];
+
 	u8         flex_parser_5[0x20];
+
 	u8         flex_parser_4[0x20];
 };
 
-struct mlx5_ifc_ste_flex_parser_tnl_bits {
-	u8         flex_parser_tunneling_header_63_32[0x20];
-	u8         flex_parser_tunneling_header_31_0[0x20];
+struct mlx5_ifc_ste_flex_parser_tnl_vxlan_gpe_bits {
+	u8         outer_vxlan_gpe_flags[0x8];
+	u8         reserved_at_8[0x10];
+	u8         outer_vxlan_gpe_next_protocol[0x8];
+
+	u8         outer_vxlan_gpe_vni[0x18];
+	u8         reserved_at_38[0x8];
+
+	u8         reserved_at_40[0x40];
+};
+
+struct mlx5_ifc_ste_flex_parser_tnl_geneve_bits {
+	u8         reserved_at_0[0x2];
+	u8         geneve_opt_len[0x6];
+	u8         geneve_oam[0x1];
+	u8         reserved_at_9[0x7];
+	u8         geneve_protocol_type[0x10];
+
+	u8         geneve_vni[0x18];
+	u8         reserved_at_38[0x8];
+
+	u8         reserved_at_40[0x40];
+};
+
+struct mlx5_ifc_ste_flex_parser_tnl_gtpu_bits {
+	u8         reserved_at_0[0x5];
+	u8         gtpu_flags[0x3];
+	u8         gtpu_msg_type[0x8];
+	u8         reserved_at_10[0x10];
+
+	u8         gtpu_teid[0x20];
+
 	u8         reserved_at_40[0x40];
 };
 
 struct mlx5_ifc_ste_general_purpose_bits {
 	u8         general_purpose_lookup_field[0x20];
+
 	u8         reserved_at_20[0x20];
+
 	u8         reserved_at_40[0x20];
+
 	u8         reserved_at_60[0x20];
 };
 
@@ -1512,12 +1665,15 @@ struct mlx5_ifc_ste_src_gvmi_qp_bits {
 	u8         loopback_syndrome[0x8];
 	u8         reserved_at_8[0x8];
 	u8         source_gvmi[0x10];
+
 	u8         reserved_at_20[0x5];
 	u8         force_lb[0x1];
 	u8         functional_lb[0x1];
 	u8         source_is_requestor[0x1];
 	u8         source_qp[0x18];
+
 	u8         reserved_at_40[0x20];
+
 	u8         reserved_at_60[0x20];
 };
 
@@ -1540,9 +1696,25 @@ struct mlx5_ifc_add_action_in_bits {
 	u8         data[0x20];
 };
 
+struct mlx5_ifc_copy_action_in_bits {
+	u8         action_type[0x4];
+	u8         src_field[0xc];
+	u8         reserved_at_10[0x3];
+	u8         src_offset[0x5];
+	u8         reserved_at_18[0x3];
+	u8         length[0x5];
+
+	u8         reserved_at_20[0x4];
+	u8         dst_field[0xc];
+	u8         reserved_at_30[0x3];
+	u8         dst_offset[0x5];
+	u8         reserved_at_38[0x8];
+};
+
 enum {
 	MLX5_ACTION_TYPE_SET   = 0x1,
 	MLX5_ACTION_TYPE_ADD   = 0x2,
+	MLX5_ACTION_TYPE_COPY  = 0x3,
 };
 
 enum {
@@ -1568,6 +1740,7 @@ enum {
 	MLX5_ACTION_IN_FIELD_OUT_DIPV6_31_0    = 0x14,
 	MLX5_ACTION_IN_FIELD_OUT_SIPV4         = 0x15,
 	MLX5_ACTION_IN_FIELD_OUT_DIPV4         = 0x16,
+	MLX5_ACTION_IN_FIELD_OUT_FIRST_VID     = 0x17,
 	MLX5_ACTION_IN_FIELD_OUT_IPV6_HOPLIMIT = 0x47,
 	MLX5_ACTION_IN_FIELD_OUT_METADATA_REGA = 0x49,
 	MLX5_ACTION_IN_FIELD_OUT_METADATA_REGB = 0x50,
@@ -1693,6 +1866,38 @@ struct mlx5_ifc_general_obj_out_cmd_hdr_bits {
 	u8         reserved_at_60[0x20];
 };
 
+struct mlx5_ifc_flow_meter_bits {
+	u8         modify_field_select[0x40];
+
+	u8         active[0x1];
+	u8         reserved_at_41[0x3];
+	u8         return_reg_id[0x4];
+	u8         table_type[0x8];
+	u8         reserved_at_50[0x10];
+
+	u8         reserved_at_60[0x8];
+	u8         destination_table_id[0x18];
+
+	u8         reserved_at_80[0x80];
+
+	u8         flow_meter_params[0x100];
+
+	u8         reserved_at_180[0x180];
+
+	u8         sw_steering_icm_address_rx[0x40];
+	u8         sw_steering_icm_address_tx[0x40];
+};
+
+struct mlx5_ifc_create_flow_meter_in_bits {
+	struct mlx5_ifc_general_obj_in_cmd_hdr_bits   hdr;
+	struct mlx5_ifc_flow_meter_bits               meter;
+};
+
+struct mlx5_ifc_query_flow_meter_out_bits {
+	struct mlx5_ifc_general_obj_out_cmd_hdr_bits   hdr;
+	struct mlx5_ifc_flow_meter_bits                obj;
+};
+
 struct mlx5_ifc_esw_vport_context_bits {
 	u8         reserved_at_0[0x3];
 	u8         vport_svlan_strip[0x1];
@@ -2108,9 +2313,25 @@ struct mlx5_ifc_dr_action_hw_set_bits {
 	u8         destination_left_shifter[0x6];
 	u8         reserved_at_18[0x3];
 	u8         destination_length[0x5];
+
 	u8         inline_data[0x20];
 };
 
+struct mlx5_ifc_dr_action_hw_copy_bits {
+	u8         opcode[0x8];
+	u8         destination_field_code[0x8];
+	u8         reserved_at_10[0x2];
+	u8         destination_left_shifter[0x6];
+	u8         reserved_at_18[0x2];
+	u8         destination_length[0x6];
+
+	u8         reserved_at_20[0x8];
+	u8         source_field_code[0x8];
+	u8         reserved_at_30[0x2];
+	u8         source_left_shifter[0x6];
+	u8         reserved_at_38[0x8];
+};
+
 enum {
 	MLX5_DR_ACTION_MDFY_HW_FLD_L2_0		= 0,
 	MLX5_DR_ACTION_MDFY_HW_FLD_L2_1		= 1,
@@ -2139,6 +2360,7 @@ enum {
 };
 
 enum {
+	MLX5_DR_ACTION_MDFY_HW_OP_COPY		= 0x1,
 	MLX5_DR_ACTION_MDFY_HW_OP_SET		= 0x2,
 	MLX5_DR_ACTION_MDFY_HW_OP_ADD		= 0x3,
 };
