@@ -46,11 +46,11 @@
 #include <stdbool.h>
 
 #include "qelr.h"
-#include "qelr_abi.h"
 #include "qelr_chain.h"
 #include "qelr_verbs.h"
 #include <util/compiler.h>
-
+#include <util/util.h>
+#include <util/mmio.h>
 #include <stdio.h>
 #include <stdlib.h>
 
@@ -58,9 +58,6 @@
 #define QELR_RQE_ELEMENT_SIZE	(sizeof(struct rdma_rq_sge))
 #define QELR_CQE_SIZE		(sizeof(union rdma_cqe))
 
-#define IS_IWARP(_dev)		(_dev->node_type == IBV_NODE_RNIC)
-#define IS_ROCE(_dev)		(_dev->node_type == IBV_NODE_CA)
-
 static void qelr_inc_sw_cons_u16(struct qelr_qp_hwq_info *info)
 {
 	info->cons = (info->cons + 1) % info->max_wr;
@@ -194,10 +191,11 @@ int qelr_dereg_mr(struct verbs_mr *vmr)
 	if (rc)
 		return rc;
 
-	free(vmr);
-
 	DP_VERBOSE(cxt->dbg_fp, QELR_MSG_MR,
 		   "MR DERegister %p completed successfully\n", vmr);
+
+	free(vmr);
+
 	return 0;
 }
 
@@ -220,7 +218,7 @@ struct ibv_cq *qelr_create_cq(struct ibv_context *context, int cqe,
 			      int comp_vector)
 {
 	struct qelr_devctx *cxt = get_qelr_ctx(context);
-	struct qelr_create_cq_resp resp;
+	struct qelr_create_cq_resp resp = {};
 	struct qelr_create_cq cmd;
 	struct qelr_cq *cq;
 	int chain_size;
@@ -265,6 +263,27 @@ struct ibv_cq *qelr_create_cq(struct ibv_context *context, int cqe,
 		RDMA_PWM_VAL32_DATA_AGG_CMD_SHIFT;
 	cq->db_addr = cxt->db_addr + resp.db_offset;
 
+	if (resp.db_rec_addr) {
+		cq->db_rec_map = mmap(NULL, cxt->kernel_page_size, PROT_WRITE,
+				      MAP_SHARED, context->cmd_fd,
+				      resp.db_rec_addr);
+		if (cq->db_rec_map == MAP_FAILED) {
+			int errsv = errno;
+
+			DP_ERR(cxt->dbg_fp,
+			       "alloc context: doorbell rec mapping failed resp.db_rec_addr = %llx size=%d context->cmd_fd=%d errno=%d\n",
+			       resp.db_rec_addr, cxt->kernel_page_size,
+			       context->cmd_fd, errsv);
+			goto err_1;
+		}
+		cq->db_rec_addr = cq->db_rec_map;
+	} else {
+		/* Kernel doesn't support doorbell recovery. Point to dummy
+		 * location instead
+		 */
+		cq->db_rec_addr = &cxt->db_rec_addr_dummy;
+	}
+
 	/* point to the very last element, passing this we will toggle */
 	cq->toggle_cqe = qelr_chain_get_last_elem(&cq->chain);
 	cq->chain_toggle = RDMA_CQE_REQUESTER_TOGGLE_BIT_MASK;
@@ -301,11 +320,14 @@ int qelr_destroy_cq(struct ibv_cq *ibv_cq)
 	}
 
 	qelr_chain_free(&cq->chain);
-	free(cq);
+	if (cq->db_rec_map)
+		munmap(cq->db_rec_map, cxt->kernel_page_size);
 
 	DP_VERBOSE(cxt->dbg_fp, QELR_MSG_CQ,
 		   "destroy cq: successfully destroyed %p\n", cq);
 
+	free(cq);
+
 	return 0;
 }
 
@@ -532,6 +554,8 @@ static inline int qelr_create_qp_buffers(struct qelr_devctx *cxt,
 	rc = qelr_create_qp_buffers_rq(cxt, qp, attrs);
 	if (rc) {
 		qelr_chain_free_sq(qp);
+		if (qp->sq.db_rec_map)
+			munmap(qp->sq.db_rec_map, cxt->kernel_page_size);
 		return rc;
 	}
 
@@ -548,6 +572,28 @@ static inline int qelr_configure_qp_sq(struct qelr_devctx *cxt,
 	qp->sq.prod = 0;
 	qp->sq.db = cxt->db_addr + resp->sq_db_offset;
 	qp->sq.edpm_db = cxt->db_addr;
+	if (resp->sq_db_rec_addr) {
+		qp->sq.db_rec_map = mmap(NULL, cxt->kernel_page_size,
+					 PROT_WRITE, MAP_SHARED,
+					 cxt->ibv_ctx.context.cmd_fd,
+					 resp->sq_db_rec_addr);
+
+		if (qp->sq.db_rec_map == MAP_FAILED) {
+			int errsv = errno;
+
+			DP_ERR(cxt->dbg_fp,
+			       "alloc context: doorbell rec mapping failed resp.db_rec_addr = %llx size=%d context->cmd_fd=%d errno=%d\n",
+			       resp->sq_db_rec_addr, cxt->kernel_page_size,
+			       cxt->ibv_ctx.context.cmd_fd, errsv);
+			return -ENOMEM;
+		}
+		qp->sq.db_rec_addr = qp->sq.db_rec_map;
+	} else {
+		/* Kernel doesn't support doorbell recovery. Point to dummy
+		 * location instead
+		 */
+		qp->sq.db_rec_addr = &cxt->db_rec_addr_dummy;
+	}
 
 	/* shadow SQ */
 	qp->sq.max_wr++;	/* prod/cons method requires N+1 elements */
@@ -574,6 +620,28 @@ static inline int qelr_configure_qp_rq(struct qelr_devctx *cxt,
 	qp->rq.iwarp_db2_data.data.value = htole16(DQ_TCM_IWARP_POST_RQ_CF_CMD);
 	qp->rq.prod = 0;
 
+	if (resp->rq_db_rec_addr) {
+		qp->rq.db_rec_map = mmap(NULL, cxt->kernel_page_size,
+					 PROT_WRITE, MAP_SHARED,
+					 cxt->ibv_ctx.context.cmd_fd,
+					 resp->rq_db_rec_addr);
+		if (qp->rq.db_rec_map == MAP_FAILED) {
+			int errsv = errno;
+
+			DP_ERR(cxt->dbg_fp,
+			       "alloc context: doorbell rec mapping failed resp.db_rec_addr = %llx size=%d context->cmd_fd=%d errno=%d\n",
+			       resp->rq_db_rec_addr, cxt->kernel_page_size,
+			       cxt->ibv_ctx.context.cmd_fd, errsv);
+			return -ENOMEM;
+		}
+		qp->rq.db_rec_addr = qp->rq.db_rec_map;
+	} else {
+		/* Kernel doesn't support doorbell recovery. Point to dummy
+		 * location instead
+		 */
+		qp->rq.db_rec_addr = &cxt->db_rec_addr_dummy;
+	}
+
 	/* shadow RQ */
 	qp->rq.max_wr++;	/* prod/cons method requires N+1 elements */
 	qp->rqe_wr_id = calloc(qp->rq.max_wr, sizeof(*qp->rqe_wr_id));
@@ -652,7 +720,7 @@ struct ibv_qp *qelr_create_qp(struct ibv_pd *pd,
 			      struct ibv_qp_init_attr *attrs)
 {
 	struct qelr_devctx *cxt = get_qelr_ctx(pd->context);
-	struct qelr_create_qp_resp resp;
+	struct qelr_create_qp_resp resp = {};
 	struct qelr_create_qp req;
 	struct qelr_qp *qp;
 	int rc;
@@ -986,11 +1054,16 @@ int qelr_destroy_qp(struct ibv_qp *ibqp)
 	qelr_free_rq(qp);
 	qelr_chain_free_sq(qp);
 	qelr_chain_free_rq(qp);
-	free(qp);
+	if (qp->sq.db_rec_map)
+		munmap(qp->sq.db_rec_map, cxt->kernel_page_size);
+	if (qp->rq.db_rec_map)
+		munmap(qp->rq.db_rec_map, cxt->kernel_page_size);
 
 	DP_VERBOSE(cxt->dbg_fp, QELR_MSG_QP,
 		   "destroy cq: successfully destroyed %p\n", qp);
 
+	free(qp);
+
 	return 0;
 }
 
@@ -1019,16 +1092,27 @@ static inline void qelr_init_dpm_info(struct qelr_devctx *cxt,
 				      int data_size)
 {
 	dpm->is_edpm = 0;
+	dpm->is_ldpm = 0;
 
-	/* Currently dpm is not supported for iWARP */
-	if (IS_IWARP(cxt->ibv_ctx.context.device))
+	/* DPM only succeeds when transmit queues are empty */
+	if (!qelr_chain_is_full(&qp->sq.chain))
 		return;
 
-	if (qelr_chain_is_full(&qp->sq.chain) &&
-	    wr->send_flags & IBV_SEND_INLINE && !qp->edpm_disabled) {
+	/* Check if edpm can be used */
+	if (wr->send_flags & IBV_SEND_INLINE && !qp->edpm_disabled &&
+	    cxt->dpm_flags & QELR_DPM_FLAGS_ENHANCED) {
 		memset(dpm, 0, sizeof(*dpm));
 		dpm->rdma_ext = (struct qelr_rdma_ext *)&dpm->payload;
 		dpm->is_edpm = 1;
+		return;
+	}
+
+	 /* Check if ldpm can be used - not inline and limited to ldpm_limit */
+	if (cxt->dpm_flags & QELR_DPM_FLAGS_LEGACY &&
+	    !(wr->send_flags & IBV_SEND_INLINE) &&
+	    data_size <= cxt->ldpm_limit_size) {
+		memset(dpm, 0, sizeof(*dpm));
+		dpm->is_ldpm = 1;
 	}
 }
 
@@ -1036,9 +1120,11 @@ static inline void qelr_init_dpm_info(struct qelr_devctx *cxt,
 #define QELR_IB_OPCODE_SEND_ONLY_WITH_IMMEDIATE          0x05
 #define QELR_IB_OPCODE_RDMA_WRITE_ONLY                   0x0a
 #define QELR_IB_OPCODE_RDMA_WRITE_ONLY_WITH_IMMEDIATE    0x0b
-#define QELR_IS_IMM(opcode) \
-	((opcode == QELR_IB_OPCODE_SEND_ONLY_WITH_IMMEDIATE) || \
-	 (opcode == QELR_IB_OPCODE_RDMA_WRITE_ONLY_WITH_IMMEDIATE))
+#define QELR_IB_OPCODE_SEND_WITH_INV			 0x17
+#define QELR_IS_IMM_OR_INV(opcode) \
+	(((opcode) == QELR_IB_OPCODE_SEND_ONLY_WITH_IMMEDIATE) || \
+	 ((opcode) == QELR_IB_OPCODE_RDMA_WRITE_ONLY_WITH_IMMEDIATE) || \
+	 ((opcode) == QELR_IB_OPCODE_SEND_WITH_INV))
 
 static inline void qelr_edpm_set_msg_data(struct qelr_qp *qp,
 					  struct qelr_dpm *dpm,
@@ -1050,7 +1136,7 @@ static inline void qelr_edpm_set_msg_data(struct qelr_qp *qp,
 	uint32_t wqe_size, dpm_size, params;
 
 	params = 0;
-	wqe_size = length + (QELR_IS_IMM(opcode)? sizeof(uint32_t) : 0);
+	wqe_size = length + (QELR_IS_IMM_OR_INV(opcode) ? sizeof(uint32_t) : 0);
 	dpm_size = wqe_size + sizeof(struct db_roce_dpm_data);
 
 	SET_FIELD(params, DB_ROCE_DPM_PARAMS_DPM_TYPE, DPM_ROCE);
@@ -1178,6 +1264,12 @@ static void qelr_prepare_sq_sges(struct qelr_qp *qp,
 		TYPEPTR_ADDR_SET(sge, addr, wr->sg_list[i].addr);
 		sge->l_key = htole32(wr->sg_list[i].lkey);
 		sge->length = htole32(wr->sg_list[i].length);
+
+		if (dpm->is_ldpm) {
+			memcpy(&dpm->payload[dpm->payload_size], sge,
+			       sizeof(*sge));
+			dpm->payload_size += sizeof(*sge);
+		}
 	}
 
 	if (wqe_size)
@@ -1212,8 +1304,16 @@ static uint32_t qelr_prepare_sq_rdma_data(struct qelr_qp *qp,
 					    &rwqe->flags, flags);
 		rwqe->wqe_size = *p_wqe_size;
 	} else {
+		if (dpm->is_ldpm)
+			dpm->payload_size = sizeof(*rwqe) + sizeof(*rwqe2);
 		qelr_prepare_sq_sges(qp, dpm, p_wqe_size, wr);
 		rwqe->wqe_size = *p_wqe_size;
+
+		if (dpm->is_ldpm) {
+			memcpy(dpm->payload, rwqe, sizeof(*rwqe));
+			memcpy(&dpm->payload[sizeof(*rwqe)], rwqe2,
+			       sizeof(*rwqe2));
+		}
 	}
 
 	return data_size;
@@ -1243,13 +1343,56 @@ static uint32_t qelr_prepare_sq_send_data(struct qelr_qp *qp,
 					    &swqe->flags, flags);
 		swqe->wqe_size = *p_wqe_size;
 	} else {
+		if (dpm->is_ldpm)
+			dpm->payload_size = sizeof(*swqe) + sizeof(*swqe2);
+
 		qelr_prepare_sq_sges(qp, dpm, p_wqe_size, wr);
 		swqe->wqe_size = *p_wqe_size;
+		if (dpm->is_ldpm) {
+			memcpy(dpm->payload, swqe, sizeof(*swqe));
+			memcpy(&dpm->payload[sizeof(*swqe)], swqe2,
+			       sizeof(*swqe2));
+		}
 	}
 
 	return data_size;
 }
 
+static void qelr_prepare_sq_atom_data(struct qelr_qp *qp,
+				      struct qelr_dpm *dpm,
+				      struct rdma_sq_atomic_wqe_1st *awqe1,
+				      struct rdma_sq_atomic_wqe_2nd *awqe2,
+				      struct rdma_sq_atomic_wqe_3rd *awqe3,
+				      struct ibv_send_wr *wr)
+{
+	if (dpm->is_ldpm) {
+		memcpy(&dpm->payload[dpm->payload_size], awqe1, sizeof(*awqe1));
+		dpm->payload_size += sizeof(*awqe1);
+		memcpy(&dpm->payload[dpm->payload_size], awqe2, sizeof(*awqe2));
+		dpm->payload_size += sizeof(*awqe2);
+		memcpy(&dpm->payload[dpm->payload_size], awqe3, sizeof(*awqe3));
+		dpm->payload_size += sizeof(*awqe3);
+	}
+
+	qelr_prepare_sq_sges(qp, dpm, NULL, wr);
+}
+
+static inline void qelr_ldpm_prepare_data(struct qelr_qp *qp,
+					  struct qelr_dpm *dpm)
+{
+	uint32_t val, params;
+
+	/* DPM size is given in 8 bytes so we round up */
+	val = dpm->payload_size + sizeof(struct db_roce_dpm_data);
+	val = DIV_ROUND_UP(val, sizeof(uint64_t));
+
+	params = 0;
+	SET_FIELD(params, DB_ROCE_DPM_PARAMS_SIZE, val);
+	SET_FIELD(params, DB_ROCE_DPM_PARAMS_DPM_TYPE, DPM_LEGACY);
+
+	dpm->msg.data.params.params = htole32(params);
+}
+
 static enum ibv_wc_opcode qelr_ibv_to_wc_opcode(enum ibv_wr_opcode opcode)
 {
 	switch (opcode) {
@@ -1258,6 +1401,7 @@ static enum ibv_wc_opcode qelr_ibv_to_wc_opcode(enum ibv_wr_opcode opcode)
 		return IBV_WC_RDMA_WRITE;
 	case IBV_WR_SEND_WITH_IMM:
 	case IBV_WR_SEND:
+	case IBV_WR_SEND_WITH_INV:
 		return IBV_WC_SEND;
 	case IBV_WR_RDMA_READ:
 		return IBV_WC_RDMA_READ;
@@ -1274,6 +1418,8 @@ static inline void doorbell_qp(struct qelr_qp *qp)
 {
 	mmio_wc_start();
 	writel(qp->sq.db_data.raw, qp->sq.db);
+	/* copy value to doorbell recovery mechanism */
+	qp->sq.db_rec_addr->db_data = qp->sq.db_data.raw;
 	mmio_flush_writes();
 }
 
@@ -1281,7 +1427,6 @@ static inline void doorbell_dpm_qp(struct qelr_devctx *cxt, struct qelr_qp *qp,
 				   struct qelr_dpm *dpm)
 {
 	uint32_t offset = 0;
-	uint64_t data;
 	uint64_t *payload = (uint64_t *)dpm->payload;
 	uint32_t num_dwords;
 	int bytes = 0;
@@ -1295,25 +1440,32 @@ static inline void doorbell_dpm_qp(struct qelr_devctx *cxt, struct qelr_qp *qp,
 	db_addr = qp->sq.edpm_db;
 	writeq(dpm->msg.raw, db_addr);
 
-
 	/* Write mesage body */
 	bytes += sizeof(uint64_t);
-	num_dwords = (dpm->payload_size + sizeof(uint64_t) - 1) /
-		sizeof(uint64_t);
+	num_dwords = DIV_ROUND_UP(dpm->payload_size, sizeof(uint64_t));
+
 	db_addr += sizeof(dpm->msg.data);
 
+	if (bytes == cxt->edpm_trans_size) {
+		mmio_flush_writes();
+		bytes = 0;
+	}
+
 	while (offset < num_dwords) {
-		data = payload[offset];
-		writeq(data, db_addr);
+		/* endianity is different between FW and DORQ HW block */
+		if (dpm->is_ldpm)
+			mmio_write64_be(db_addr, htobe64(payload[offset]));
+		else /* EDPM */
+			mmio_write64(db_addr, payload[offset]);
 
 		bytes += sizeof(uint64_t);
 		db_addr += sizeof(uint64_t);
 
-		/* Since we rewrite the buffer every 64 bytes we need to flush
-		 * it here, otherwise the CPU could optimize away the
-		 * duplicate stores.
+		/* Writing to a wc bar. We need to flush the writes every
+		 * edpm transaction size otherwise the CPU could optimize away
+		 * the duplicate stores.
 		 */
-		if (bytes == 64) {
+		if (bytes == cxt->edpm_trans_size) {
 			mmio_flush_writes();
 			bytes = 0;
 		}
@@ -1427,6 +1579,8 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 			qelr_edpm_set_msg_data(qp, &dpm,
 					       QELR_IB_OPCODE_SEND_ONLY_WITH_IMMEDIATE,
 					       wqe_length, se, comp);
+		else if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
 
 		qp->wqe_wr_id[qp->sq.prod].wqe_size = wqe_size;
 		qp->prev_wqe_size = wqe_size;
@@ -1442,17 +1596,48 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 		wqe_length = qelr_prepare_sq_send_data(qp, &dpm, data_size,
 						       &wqe_size, swqe, swqe2,
 						       wr, 0);
-
 		if (dpm.is_edpm)
 			qelr_edpm_set_msg_data(qp, &dpm,
 					       QELR_IB_OPCODE_SEND_ONLY,
 					       wqe_length, se, comp);
+		else if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
 
 		qp->wqe_wr_id[qp->sq.prod].wqe_size = wqe_size;
 		qp->prev_wqe_size = wqe_size;
 		qp->wqe_wr_id[qp->sq.prod].bytes_len = wqe_length;
 		break;
 
+	case IBV_WR_SEND_WITH_INV:
+		wqe->req_type = RDMA_SQ_REQ_TYPE_SEND_WITH_INVALIDATE;
+		swqe = (struct rdma_sq_send_wqe_1st *)wqe;
+
+		wqe_size = sizeof(struct rdma_sq_send_wqe) / RDMA_WQE_BYTES;
+		swqe2 = qelr_chain_produce(&qp->sq.chain);
+
+		if (dpm.is_edpm)
+			qelr_edpm_set_inv_imm(qp, &dpm,
+					      htobe32(wr->invalidate_rkey));
+
+		swqe->inv_key_or_imm_data = htole32(wr->invalidate_rkey);
+
+		wqe_length = qelr_prepare_sq_send_data(qp, &dpm, data_size,
+						       &wqe_size, swqe, swqe2,
+						       wr, 0);
+
+		if (dpm.is_edpm)
+			qelr_edpm_set_msg_data(qp, &dpm,
+					       QELR_IB_OPCODE_SEND_WITH_INV,
+					       wqe_length, se, comp);
+		else if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
+
+		qp->wqe_wr_id[qp->sq.prod].wqe_size = wqe_size;
+		qp->prev_wqe_size = wqe_size;
+		qp->wqe_wr_id[qp->sq.prod].bytes_len = wqe_length;
+
+		break;
+
 	case IBV_WR_RDMA_WRITE_WITH_IMM:
 		wqe->req_type = RDMA_SQ_REQ_TYPE_RDMA_WR_WITH_IMM;
 		rwqe = (struct rdma_sq_rdma_wqe_1st *)wqe;
@@ -1472,6 +1657,9 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 					       QELR_IB_OPCODE_RDMA_WRITE_ONLY_WITH_IMMEDIATE,
 					       wqe_length + sizeof(*dpm.rdma_ext),
 					       se, comp);
+		else if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
+
 		qp->wqe_wr_id[qp->sq.prod].wqe_size = wqe_size;
 		qp->prev_wqe_size = wqe_size;
 		qp->wqe_wr_id[qp->sq.prod].bytes_len = wqe_length;
@@ -1496,6 +1684,9 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 					       wqe_length +
 					       sizeof(*dpm.rdma_ext),
 					       se, comp);
+		else if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
+
 		qp->wqe_wr_id[qp->sq.prod].wqe_size = wqe_size;
 		qp->prev_wqe_size = wqe_size;
 		qp->wqe_wr_id[qp->sq.prod].bytes_len = wqe_length;
@@ -1509,6 +1700,9 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 		rwqe2 = (struct rdma_sq_rdma_wqe_2nd *)qelr_chain_produce(&qp->sq.chain);
 		wqe_length = qelr_prepare_sq_rdma_data(qp, &dpm, data_size, &wqe_size,
 						       rwqe, rwqe2, wr, 0);
+		if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
+
 		qp->wqe_wr_id[qp->sq.prod].wqe_size = wqe_size;
 		qp->prev_wqe_size = wqe_size;
 		qp->wqe_wr_id[qp->sq.prod].bytes_len = wqe_length;
@@ -1534,8 +1728,9 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 			TYPEPTR_ADDR_SET(awqe3, cmp_data, wr->wr.atomic.compare_add);
 		}
 
-		qelr_prepare_sq_sges(qp, &dpm, NULL, wr);
-
+		qelr_prepare_sq_atom_data(qp, &dpm, awqe1, awqe2, awqe3, wr);
+		if (dpm.is_ldpm)
+			qelr_ldpm_prepare_data(qp, &dpm);
 		qp->wqe_wr_id[qp->sq.prod].wqe_size = awqe1->wqe_size;
 		qp->prev_wqe_size = awqe1->wqe_size;
 
@@ -1565,7 +1760,7 @@ static int __qelr_post_send(struct qelr_devctx *cxt, struct qelr_qp *qp,
 	db_val = le16toh(qp->sq.db_data.data.value) + 1;
 	qp->sq.db_data.data.value = htole16(db_val);
 
-	if (dpm.is_edpm) {
+	if (dpm.is_edpm || dpm.is_ldpm) {
 		doorbell_dpm_qp(cxt, qp, &dpm);
 		*normal_db_required = 0;
 	} else {
@@ -1797,6 +1992,8 @@ int qelr_post_recv(struct ibv_qp *ibqp, struct ibv_recv_wr *wr,
 		qp->rq.db_data.data.value = htole16(db_val);
 
 		writel(qp->rq.db_data.raw, qp->rq.db);
+		/* copy value to doorbell recovery mechanism */
+		qp->rq.db_rec_addr->db_data = qp->rq.db_data.raw;
 		mmio_flush_writes();
 
 		if (iwarp) {
@@ -2006,7 +2203,6 @@ static void __process_resp_one(struct qelr_qp *qp, struct qelr_cq *cq,
 	wc->opcode = IBV_WC_RECV;
 	wc->wr_id = wr_id;
 	wc->wc_flags = 0;
-
 	switch (resp->status) {
 	case RDMA_CQE_RESP_STS_LOCAL_ACCESS_ERR:
 		wc_status = IBV_WC_LOC_ACCESS_ERR;
@@ -2041,6 +2237,10 @@ static void __process_resp_one(struct qelr_qp *qp, struct qelr_cq *cq,
 			wc->imm_data = htobe32(le32toh(resp->imm_data_or_inv_r_Key));
 			wc->wc_flags |= IBV_WC_WITH_IMM;
 			break;
+		case QELR_RESP_INV:
+			wc->invalidated_rkey = le32toh(resp->imm_data_or_inv_r_Key);
+			wc->wc_flags |= IBV_WC_WITH_INV;
+			break;
 		case QELR_RESP_RDMA:
 			DP_ERR(cxt->dbg_fp, "Invalid flags detected\n");
 			break;
@@ -2188,6 +2388,8 @@ static void doorbell_cq(struct qelr_cq *cq, uint32_t cons, uint8_t flags)
 	cq->db.data.value = htole32(cons);
 
 	writeq(cq->db.raw, cq->db_addr);
+	/* copy value to doorbell recovery mechanism */
+	cq->db_rec_addr->db_data = cq->db.raw;
 	mmio_flush_writes();
 }
 
