@@ -137,8 +137,8 @@ struct ibv_srq *pvrdma_create_srq(struct ibv_pd *pd,
 	pvrdma_init_srq_queue(srq);
 
 	memset(&cmd, 0, sizeof(cmd));
-	cmd.udata.buf_addr = (uintptr_t) srq->buf.buf;
-	cmd.udata.buf_size = srq->buf.length;
+	cmd.buf_addr = (uintptr_t) srq->buf.buf;
+	cmd.buf_size = srq->buf.length;
 
 	ret = ibv_cmd_create_srq(pd, &srq->ibv_srq, attr,
 				 &cmd.ibv_cmd, sizeof(cmd),
@@ -147,7 +147,7 @@ struct ibv_srq *pvrdma_create_srq(struct ibv_pd *pd,
 	if (ret)
 		goto err_free;
 
-	srq->srqn = resp.udata.srqn;
+	srq->srqn = resp.srqn;
 
 	return &srq->ibv_srq;
 
@@ -211,9 +211,8 @@ struct ibv_qp *pvrdma_create_qp(struct ibv_pd *pd,
 {
 	struct pvrdma_device *dev = to_vdev(pd->context->device);
 	struct user_pvrdma_create_qp cmd;
-	struct ib_uverbs_create_qp_resp resp;
+	struct user_pvrdma_create_qp_resp resp = {};
 	struct pvrdma_qp *qp;
-	int ret;
 	int is_srq = !!(attr->srq);
 
 	attr->cap.max_send_sge = max_t(uint32_t, 1U, attr->cap.max_send_sge);
@@ -276,20 +275,22 @@ struct ibv_qp *pvrdma_create_qp(struct ibv_pd *pd,
 	pvrdma_init_qp_queue(qp);
 
 	memset(&cmd, 0, sizeof(cmd));
-	cmd.udata.sbuf_addr = (uintptr_t)qp->sbuf.buf;
-	cmd.udata.sbuf_size = qp->sbuf.length;
-	cmd.udata.rbuf_addr = (uintptr_t)qp->rbuf.buf;
-	cmd.udata.rbuf_size = qp->rbuf.length;
-	cmd.udata.qp_addr = (uintptr_t) qp;
+	cmd.sbuf_addr = (uintptr_t)qp->sbuf.buf;
+	cmd.sbuf_size = qp->sbuf.length;
+	cmd.rbuf_addr = (uintptr_t)qp->rbuf.buf;
+	cmd.rbuf_size = qp->rbuf.length;
+	cmd.qp_addr = (uintptr_t) qp;
 
-	ret = ibv_cmd_create_qp(pd, &qp->ibv_qp, attr,
-				&cmd.ibv_cmd, sizeof(cmd),
-				&resp, sizeof(resp));
-
-	if (ret)
+	if (ibv_cmd_create_qp(pd, &qp->ibv_qp, attr, &cmd.ibv_cmd, sizeof(cmd),
+			      &resp.ibv_resp, sizeof(resp)))
 		goto err_free;
 
-	to_vctx(pd->context)->qp_tbl[qp->ibv_qp.qp_num & 0xFFFF] = qp;
+	if (resp.drv_payload.qp_handle != 0)
+		qp->qp_handle = resp.drv_payload.qp_handle;
+	else
+		qp->qp_handle = qp->ibv_qp.qp_num;
+
+	to_vctx(pd->context)->qp_tbl[qp->qp_handle & 0xFFFF] = qp;
 
 	/* If set, each WR submitted to the SQ generate a completion entry */
 	if (attr->sq_sig_all)
@@ -313,8 +314,8 @@ err:
 }
 
 int pvrdma_query_qp(struct ibv_qp *ibqp, struct ibv_qp_attr *attr,
-		      int attr_mask,
-		      struct ibv_qp_init_attr *init_attr)
+		    int attr_mask,
+		    struct ibv_qp_init_attr *init_attr)
 {
 	struct ibv_query_qp cmd;
 	struct pvrdma_qp *qp = to_vqp(ibqp);
@@ -335,25 +336,26 @@ int pvrdma_query_qp(struct ibv_qp *ibqp, struct ibv_qp_attr *attr,
 	return 0;
 }
 
-int pvrdma_modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
-		       int attr_mask)
+int pvrdma_modify_qp(struct ibv_qp *ibqp, struct ibv_qp_attr *attr,
+		     int attr_mask)
 {
 	struct ibv_modify_qp cmd;
+	struct pvrdma_qp *qp = to_vqp(ibqp);
 	int ret;
 
 	/* Sanity check */
 	if (!attr_mask)
 		return 0;
 
-	ret = ibv_cmd_modify_qp(qp, attr, attr_mask, &cmd, sizeof(cmd));
+	ret = ibv_cmd_modify_qp(ibqp, attr, attr_mask, &cmd, sizeof(cmd));
 
 	if (!ret &&
 	    (attr_mask & IBV_QP_STATE) &&
 	    attr->qp_state == IBV_QPS_RESET) {
-		pvrdma_cq_clean(to_vcq(qp->recv_cq), qp->qp_num);
-		if (qp->send_cq != qp->recv_cq)
-			pvrdma_cq_clean(to_vcq(qp->send_cq), qp->qp_num);
-		pvrdma_init_qp_queue(to_vqp(qp));
+		pvrdma_cq_clean(to_vcq(ibqp->recv_cq), qp->qp_handle);
+		if (ibqp->send_cq != ibqp->recv_cq)
+			pvrdma_cq_clean(to_vcq(ibqp->send_cq), qp->qp_handle);
+		pvrdma_init_qp_queue(qp);
 	}
 
 	return ret;
@@ -404,17 +406,17 @@ int pvrdma_destroy_qp(struct ibv_qp *ibqp)
 
 	pvrdma_lock_cqs(ibqp);
 	/* Dump cqs */
-	pvrdma_cq_clean_int(to_vcq(ibqp->recv_cq), ibqp->qp_num);
+	pvrdma_cq_clean_int(to_vcq(ibqp->recv_cq), qp->qp_handle);
 
 	if (ibqp->send_cq != ibqp->recv_cq)
-		pvrdma_cq_clean_int(to_vcq(ibqp->send_cq), ibqp->qp_num);
+		pvrdma_cq_clean_int(to_vcq(ibqp->send_cq), qp->qp_handle);
 	pvrdma_unlock_cqs(ibqp);
 
 	free(qp->sq.wrid);
 	free(qp->rq.wrid);
 	pvrdma_free_buf(&qp->rbuf);
 	pvrdma_free_buf(&qp->sbuf);
-	ctx->qp_tbl[ibqp->qp_num & 0xFFFF] = NULL;
+	ctx->qp_tbl[qp->qp_handle & 0xFFFF] = NULL;
 	free(qp);
 
 	return 0;
@@ -547,7 +549,7 @@ out:
 	if (nreq) {
 		udma_to_device_barrier();
 		pvrdma_write_uar_qp(ctx->uar,
-				    PVRDMA_UAR_QP_SEND | ibqp->qp_num);
+				    PVRDMA_UAR_QP_SEND | qp->qp_handle);
 	}
 
 	pthread_spin_unlock(&qp->sq.lock);
@@ -630,7 +632,7 @@ int pvrdma_post_recv(struct ibv_qp *ibqp, struct ibv_recv_wr *wr,
 out:
 	if (nreq)
 		pvrdma_write_uar_qp(ctx->uar,
-				    PVRDMA_UAR_QP_RECV | ibqp->qp_num);
+				    PVRDMA_UAR_QP_RECV | qp->qp_handle);
 
 	pthread_spin_unlock(&qp->rq.lock);
 	return ret;
